"""
Stochastic Simulator

This performs iterative scenario analysis using stochastic simulation.  Only total scores are accessible,
no aggregation or grouping is performed.

The simulator accepts one or more cases, and one or more simulation variants. Each variant is defined by
a set of parameters (fragments) whose exchange values vary according to a specification.  The variants can
either be "simple" (in which case the user specified the parameter values explicitly) or "stochastic" (in which
the parameter values are drawn from a distribution).
"""
from .parameter import LognormalParameter, UniformParameter, DirichletParameters
from collections import defaultdict
from math import floor, ceil
from pandas import DataFrame
from antelope_foreground.fragment_flows import frag_flow_lcia


class StochasticSimulator:
    def _scenario_spec(self, case):
        return tuple((*self._common_scenarios, *self._cases[case]))

    def _variant_spec(self, case, i):
        return tuple((*self._scenario_spec(case), i))

    def __init__(self, model, *common_scenarios):
        self._model = model

        self._common_scenarios = set()

        for c in common_scenarios:
            self.add_common_scenario(c)

        self._c = list()
        self._cases = defaultdict(set)
        self._v = list()
        self._variants = dict()
        self._simple = set()
        self._params = defaultdict(list)

        self._quantities = list()

        self._baselines = dict()  # case, quantity -> LciaResult
        self._results = dict()  # case, variant, quantity -> list of floats

    @property
    def cases(self):
        for c in self._c:
            yield c

    @property
    def variants(self):
        for v in self._v:
            yield v

    @property
    def quantities(self):
        for q in self._quantities:
            yield q

    def add_common_scenario(self, scenario):
        if isinstance(scenario, tuple):
            for sc in scenario:
                self._common_scenarios.add(sc)
        else:
            self._common_scenarios.add(scenario)

    def add_case(self, case, *case_params):
        if case not in self._c:
            self._c.append(case)
        for p in case_params:
            self._cases[case].add(p)

    def _check_match(self, variant, param):
        for p in self._params[variant]:
            if p.match(param):
                raise KeyError('Parameter %5.5s already specified' % param.uuid)

    def add_stochastic_variant(self, variant, key, n):
        """

        :param variant:
        :param key: a float value to index the observations
        :param n: the number of observations to conduct.
        :return:
        """
        if variant not in self._v:
            self._v.append(variant)
        self._variants[variant] = (key, n)

    def add_simple_variant(self, variant, key):
        if variant not in self._v:
            self._v.append(variant)
        self._simple.add(variant)
        self._variants[variant] = (key, 0)

    def add_uniform_parameter(self, variant, param, low, high):
        self._check_match(variant, param)
        self._params[variant].append(UniformParameter(param, low, high))

    def add_lognormal_parameter(self, variant, param, shape):
        self._check_match(variant, param)
        self._params[variant].append(LognormalParameter(param, shape))

    def add_dirichlet_parameters(self, variant, *fragments, scale=1.0, alpha=None):
        fs = list(fragments)
        for f in fs:
            self._check_match(variant, f)
        self._params[variant].append(DirichletParameters(*fs, scale=scale, alpha=alpha))

    def add_simple_parameter(self, variant, fragment, *values, mult=True, scenario=None):
        """
        In a simple variant, parameter values are specified directly by the modeler rather than from a stochastic
        process.  The number of "simulations" to run equals the maximum number of parameter values provided.
        :param variant:
        :param fragment:
        :param values:
        :param mult: whether the values are multiplicative or absolue
        :param scenario: for multiplicative params only, the scenario to use as baseline
        :return:
        """
        if variant not in self._simple:
            raise KeyError('Not a simple variant: %s' % variant)
        vals = list(values)
        key, n = self._variants[variant]
        if len(vals) > n:
            n = len(vals)
            self._variants[variant] = (key, n)
        if mult:
            if scenario is None:
                baseline = fragment.observed_ev
            else:
                baseline = fragment.exchange_value(scenario)
        else:
            baseline = 1.0

        for i, v in enumerate(vals):
            val = v * baseline
            fragment.observe(val, scenario=key+i)

    def observe_parameters(self, variant):
        key, n = self._variants[variant]
        for param in self._params[variant]:
            for i in range(n):
                param.observe(key+i)

    def run_baseline(self, quantity):
        if quantity not in self._quantities:
            self._quantities.append(quantity)
        for case in self.cases:
            sc = self._scenario_spec(case)
            r = self._model.fragment_lcia(quantity, scenario=sc)
            self._baselines[case, quantity] = r

    def run_variant(self, variant):
        for case in self.cases:
            self.run_case_variant(case, variant)

    def run_all(self):
        for variant in self.variants:
            self.run_variant(variant)

    def run_case_variant(self, case, variant):
        key, n = self._variants[variant]
        # reset scores
        for quantity in self.quantities:
            self._results[case, variant, quantity] = []

        # compute scores
        for i in range(n):
            ffs = self._model.traverse(scenario=self._variant_spec(case, key+i))
            for quantity in self.quantities:
                self._results[case, variant, quantity].append(frag_flow_lcia(ffs, quantity).total())
        print('%s {%s}: completed %d simulations' % (case, variant, n))

    def _5_95_ci(self, case, variant, quantity):
        r = self._results[case, variant, quantity]
        _ss = sorted(r)
        _ln = len(_ss)
        return _ss[int(floor(0.05*_ln))], _ss[int(ceil(0.95*_ln)) - 1]

    def baseline(self, case, quantity):
        return self._baselines[case, quantity]

    def result(self, case, variant, quantity):
        return self._results[case, variant, quantity]

    def output_row(self, case, variant, quantity):
        baseline = self._baselines[case, quantity].total()
        low_5, high_95 = self._5_95_ci(case, variant, quantity)
        return {
            'Case': case,
            'Variant': variant,
            'Quantity': quantity['ShortName'],
            'Indicator': quantity['Indicator'],
            'Unit': quantity.unit,
            'Baseline': baseline,
            'Low_5': low_5,
            'High_95': high_95
        }

    def output_table(self):
        return DataFrame((self.output_row(c, v, q) for c in self.cases for v in self.variants for q in self.quantities))



import csv

from collections import defaultdict
from pandas import DataFrame, MultiIndex

from antelope_core.lcia_results import LciaResult


def tabularx_ify(df, filename, width='\\textwidth', column_format='\\tabspec', hrules=True, **kwargs):
    """
    Need to figure out how to bring \tabspec in-- (answer: jinja?)
    :param df:
    :param filename:
    :param width:
    :param column_format: column format specification
    :param hrules: True
    #:param sort_column: optional integer column number by which to sort (most positive to most negative)
    # did this get dropped or what
    :return:
    """
    longstr = df.style.to_latex(column_format=column_format, hrules=hrules, **kwargs)  # df.style.to_latex recommended but has dependencies
    tabularx = longstr.replace(
        '{tabular}', '{tabularx}').replace(
        'begin{tabularx}', 'begin{tabularx}{%s}' % width)
    with open(filename, 'w') as fp:
        fp.write(tabularx)


def weigh_lcia_results(quantity, *args, weight=None):
    """
    Merge together a collection of LciaResult objects
    :param quantity: the weighing indicator- should be distinct from the supplied results but doesn't have to be
    :param args: a list of LciaResult objects to weight
    :param weight: optional dict of quantity-to-weight (if omitted, all results will be given unit weight)
    :return:
    """
    if weight is None:
        weight = dict()
    scenarios = sorted(set(arg.scenario for arg in args))
    if len(scenarios) > 1:
        print('Warning: multiple scenarios being combined together; using first only\n%s' % scenarios)
    scenario = scenarios[0]

    result = LciaResult(quantity, scenario=scenario)

    for arg in args:
        if arg.quantity in weight:
            _w = weight[arg.quantity]
        else:
            _w = 1.0

        for k in arg.keys():
            c = arg[k]
            result.add_summary(k, c.entity, 1.0, c.cumulative_result * _w)

    return result


class LcaModelRunner(object):
    _agg_key = None
    _alt_agg_key = None
    _seen_stages = None
    _fmt = '%.10e'

    def __init__(self, agg_key=None, alt_agg_key=None):
        """

        :param agg_key: default is StageName
        """
        self._scenarios = []  # sequential list of scenario names

        self._lcia_methods = []
        self._weightings = dict()

        self._publish = None

        self._results = dict()
        self.set_agg_key(agg_key)
        self.set_alt_agg_key(alt_agg_key)

    def recalculate(self, **kwargs):
        self._results = dict()
        for lm in self.lcia_methods:
            self.run_lcia(lm, **kwargs)
        for w in self.weightings:
            self._run_weighting(w)

    def set_publication_quantities(self, *qs):
        """
        Use this to constrain automatically-generated output to a subset of calculated quantities.
        :param qs:
        :return:
        """
        self._publish = list(q for q in qs if q in self._lcia_methods or q in self._weightings)

    def set_agg_key(self, agg_key=None):
        if agg_key is None:
            agg_key = lambda x: x.name

        self._agg_key = agg_key
        self._seen_stages = defaultdict(set)  # reset

    def set_alt_agg_key(self, alt_agg_key=None):
        self._alt_agg_key = alt_agg_key
        self._seen_stages = defaultdict(set)

    def _agg(self, entity):
        """
        returns either a single or a tuple
        :return:
        """
        if self._alt_agg_key is None:
            return self._agg_key(entity), None
        return self._agg_key(entity), self._alt_agg_key(entity)

    def _scenario_index(self, scenario):
        return scenario

    @property
    def scenarios(self):
        """
        This returns keys for the 'scenarios' in the tool (first result index)
        Must be implemented in a subclass
        :return:
        """
        for k in self._scenarios:
            yield k

    def add_scenario(self, name):
        if name in self._scenarios:
            raise KeyError('Case already exists: %s' % name)
        self._scenarios.append(name)

    @property
    def quantities(self):
        """
        This returns known quantities (second result index)
        :return:
        """
        if self._publish:
            for k in self._publish:
                yield k
        else:
            for k in self.lcia_methods:
                yield k
            for k in self.weightings:
                yield k

    @property
    def lcia_methods(self):
        """
        This returns the LCIA methods, which are a subset of quantities
        :return:
        """
        return self._lcia_methods

    @property
    def weightings(self):
        """
        This returns LCIA weightings, which are the complementary subset to lcia_methods
        :return:
        """
        return list(k for k in self._weightings.keys())

    @property
    def format(self):
        return self._fmt

    @format.setter
    def format(self, fmt):
        """
        None means do not format output; return raw numbers
        :param fmt:
        :return:
        """
        if fmt is None:
            self._fmt = None
        else:
            self._fmt = str(fmt)

    def add_weighting(self, quantity, *measures, weight=None):
        """
        Compute a weighted LCIA result
        :param quantity:
        :param measures: a list of LCIA quantities to be weighed
        :param weight: an optional dictionary of quantity: weight (default is equal weighting)
        :return:
        """
        if weight is None:
            weight = {m: 1.0 for m in measures}

        self._weightings[quantity] = weight
        self._run_weighting(quantity)

    def _run_case_weighting(self, scen, quantity):
        ws = self._weightings[quantity]
        res = [self.result(scen, q) for q in ws.keys()]
        wgt = weigh_lcia_results(quantity, *res, weight=ws)
        self._results[scen, quantity] = wgt

    def _run_weighting(self, quantity):
        ws = self._weightings[quantity]
        for q in ws.keys():
            self.run_lcia(q)
        for scen in self.scenarios:
            self._run_case_weighting(scen, quantity)

    @property
    def stages(self):
        return sorted(k for k, v in self._seen_stages.items() if len(v) > 0)

    def scenarios_with_stage(self, stage):
        return self._seen_stages[stage]

    def result(self, scenario, lcia_method):
        return self._results[scenario, lcia_method]

    def run_lcia_case_method(self, scen, lcia, **kwargs):
        res = self._run_scenario_lcia(scen, lcia, **kwargs)
        res.scenario = scen
        for stg in list(res.aggregate(key=self._agg).keys()):
            self._seen_stages[stg].add(scen)
        self._results[scen, lcia] = res

    def run_lcia(self, lcia, **kwargs):
        if lcia not in self._lcia_methods:
            self._lcia_methods.append(lcia)
        for scen in self.scenarios:
            self.run_lcia_case_method(scen, lcia, **kwargs)
        return self.lcia_results(lcia)

    def lcia_results(self, lcia):
        return [self._results[scenario, lcia] for scenario in self.scenarios]

    def _format(self, result):
        if self._fmt is None:
            return result
        return self._fmt % result

    def _csv_formatter(self, style):
        """
        Specifies how the rows should be constructed for the CSV result writer.
        Designed to be subclassed so that mix-ins can provide new output styles.  The proper way to do this is to
        define a method named '_csv_format_%s' % style which returns: (header_list, row_generator)
        The row_generator method should take as argument (scenario, quantity, **kwargs) and yield a dict whose keys are
        a subset of header_list and whose values are written into the file.

        Note: in practice it takes a lot of gymnastics to be able to take advantage of this (e.g. to debug /alter csv
        reports without exiting the session.
        :param style:
        :return:
        """
        st_name = '_csv_format_%s' % style
        if hasattr(self, st_name):
            return getattr(self, st_name)()
        return self.results_headings, self._gen_lcia_rows

    @property
    def results_headings(self):
        return ['scenario', 'stage', 'alt_stage', 'method', 'category', 'indicator', 'result', 'units']

    # tabular for all: accept *args as result items, go through them one by one
    def results_to_csv(self, filename, scenarios=None, style=None, aggregate=True, **kwargs):
        if scenarios is None:
            scenarios = sorted(self.scenarios)
        else:
            known = list(self.scenarios)
            scenarios = list(filter(lambda x: x in known, scenarios))

        headings, agg = self._csv_formatter(style)

        with open(filename, 'w') as fp:
            cvf = csv.DictWriter(fp, headings, quoting=csv.QUOTE_NONNUMERIC, lineterminator='\n')
            cvf.writeheader()

            for q in self.quantities:
                for scenario in scenarios:
                    for k in agg(scenario, q, aggregate=aggregate, **kwargs):
                        cvf.writerow(k)

    @staticmethod
    def _gen_row(q, k):
        k['method'] = q.get('method') or ''
        k['category'] = q.get('category') or q.name
        k['indicator'] = q['indicator']
        k['units'] = q.unit
        return k

    def _gen_lcia_rows(self, scenario, q, include_total=False, aggregate=True, **kwargs):
        res = self._results[scenario, q]
        if aggregate:
            _it = sorted(res.aggregate(key=self._agg).components(), key=lambda x: x.entity)
        else:
            _it = sorted(res.components(), key=lambda x: self._agg(x.entity))
        for c in _it:
            if aggregate:
                stage, alt_stage = c.entity
            else:
                stage, alt_stage = self._agg(c.entity)

            result = c.cumulative_result
            yield self._gen_row(q, {
                'scenario': str(scenario),
                'stage': stage,
                'alt_stage': alt_stage,
                'result': self._format(result)
            })
        if include_total:
            yield self._gen_row(q, {
                'scenario': str(scenario),
                'stage': 'Net Total',
                'result': self._format(res.total())
                 })

    def _finish_dt_output(self, dt, column_order, filename, norm=False, _total=None, add_row_index=False):
        """
        Add a units column, order it first, and transpose
        :param dt:
        :param column_order:
        :param filename:
        :param _total: whether to add scenario total to the dataframe- which should happen in client code
        :param add_row_index: for detail view only: whether scenario index should be set (perhaps always?)
        :return:
        """
        if column_order is None:
            ord_columns = list(dt.columns)
        else:
            ord_columns = [k for k in column_order if k in dt.columns]
            ord_columns += [k for k in dt.columns if k not in ord_columns]  # add in anything that's missed

        dto = dt[ord_columns].transpose()

        if add_row_index:
            dto = dto.set_index(self._scenario_index(k) for k in ord_columns)

        if norm:
            dn = DataFrame({'Normalization': [lm.norm() for lm in self.quantities]},
                           index=MultiIndex.from_tuples(self._qty_tuples)).transpose()
            dto = dn.append(dto)

        if _total:
            dt = DataFrame({'Total': [self._format(self.result(_total, lm).total()) for lm in self.quantities]},
                           index=MultiIndex.from_tuples(self._qty_tuples)).transpose()
            dto = dto.append(dt)

        if filename is not None:
            dto.to_csv(filename, quoting=csv.QUOTE_ALL)

        return dto

    @property
    def _qty_tuples(self):
        if all(q['Indicator'] == q.unit for q in self.quantities):
            for q in self.quantities:
                if q.has_property('ShortName'):
                    yield q['ShortName'], q['Indicator']
                else:
                    yield q['Name'], q['Indicator']
        else:
            for q in self.quantities:
                if q.has_property('ShortName'):
                    yield q['ShortName'], q['Indicator'], q.unit
                else:
                    yield q['Name'], q['Indicator'], q.unit

    def scenario_detail_tbl(self, scenario, filename=None, column_order=None, norm=False, total=False):
        dt = DataFrame(({k.entity: self._format(k.cumulative_result)
                         for k in self.result(scenario, lm).aggregate(key=self._agg).components()}
                        for lm in self.quantities), index=MultiIndex.from_tuples(self._qty_tuples))
        if total:
            _total = scenario
        else:
            _total = None
        return self._finish_dt_output(dt, column_order, filename, norm=norm, _total=_total)

    def scenario_summary_tbl(self, filename=None, column_order=None, norm=False, add_row_index=False):
        if column_order is None:
            column_order = list(self.scenarios)
        dt = DataFrame(({k: self._format(self.result(k, lm).total()) for k in column_order}
                        for lm in self.quantities),
                       index=MultiIndex.from_tuples(self._qty_tuples))
        return self._finish_dt_output(dt, column_order, filename, norm=norm, add_row_index=add_row_index)
    
    def results_to_tex(self, filename, scenario=None, format=None, sort_column=None, **kwargs):
        """
        Print summary table (scenario=None) or detail table (scenario is not None) in tabularx format
        :param filename: tex file
        :param scenario:
        :param format: temporarily set output format for TeX file
        :param sort_column: optional integer column number to sort by
        :param kwargs:
        :return:
        """
        oldformat = self.format

        if format:
            self.format = format

        if scenario is None:
            df = self.scenario_summary_tbl(**kwargs)
        else:
            df = self.scenario_detail_tbl(scenario, **kwargs)

        if sort_column is not None:
            # this shenanigan is necessary because tex output is often string-ified for clean formatting
            df['sort'] = df.iloc[:, sort_column].apply(float)
            try:
                df = df.sort_values('sort', ascending=False).drop('sort', axis=1)
            except ValueError:
                print('Unable to sort values~~ sorry')
                df = df.drop('sort', axis=1)

        tabularx_ify(df, filename)

        self.format = oldformat

    '''
    Subclass must implement only one function: a mapping from scenario key and lcia method to result
    '''
    def _run_scenario_lcia(self, scenario, lcia, **kwargs):
        """
        Maps scenario name to LCIA Result. Must be implemented in a subclass
        :param scenario:
        :param lcia:
        :return: LciaResult
        """
        return NotImplemented

    def to_dataframe(self, index=None, summary=True, **kwargs):
        """
        A minimal dataframe that largely duplicates scenario_summary_tbl()
        :param index:
        :param summary: [True] report totals for each case. [False] group by agg_key
        :param kwargs:
        :return:
        """
        if index is None:
            index = list(self.quantities)
        if summary:
            return DataFrame(({case: self._results[case, q].total() for case in self.scenarios}
                              for q in self.quantities),
                             index=index, **kwargs)
        else:
            return DataFrame(((q['ShortName'], q.unit, scenario, c.entity[0], c.entity[1], c.cumulative_result)
                              for scenario in self.scenarios
                              for q in self.quantities
                              for c in self.result(scenario, q).aggregate(key=self._agg).components()),
                             columns=('Quantity', 'Unit', 'Case', 'Stage', 'Alt', 'Result'))



"""
Support operations for TRACI LCIA Methods
"""
from antelope import EntityNotFound


def traci_2_replicate_nox_no2(q):
    """
    For any LCIA method, replicate factors of 'nitrogen oxides' flowable to 'nitrogen dioxide' flowable
    Applicable to USEEIO 1.1 implementation of TRACI 2.1.
    :param q:
    :return:
    """
    for cf in q.factors(flowable='nitrogen oxides'):
        q.characterize(flowable='nitrogen dioxide', ref_quantity=cf.ref_quantity, context=cf.context, value=cf.value)


def traci_2_combined_eutrophication(traci, fg, external_ref='Eutrophication', omit_n2=True):
    """
    Construct a combined eutrophication indicator that is the union of the TRACI 2.1 Eutrophication Air and
    Eutrophication Water methods.
    :param traci: Catalog query containing the TRACI 2.1 implementation
    :param fg: Foreground to contain the new combined eutrophication method
    :param external_ref: ('Eutrophication') what external reference to assign to the newly created quantity
    :param omit_n2: [True] skip the CF on gaseous nitrogen
    :return:
    """
    old_euts = [traci.get(k) for k in ('Eutrophication Air', 'Eutrophication Water')]
    try:
        return fg.get(external_ref)
    except EntityNotFound:
        pass

    new_eut = fg.new_quantity('Eutrophication Air + Water', ref_unit='kg N eq', external_ref=external_ref,
                              Method='TRACI 2.1 - Reference',
                              Category='Eutrophication', ShortName='Eutrophication', Indicator='kg N eq',
                              uuid='69726949-4add-4605-8f40-61e56f2b412c',
                              Comment="Union of TRACI 2.1 'Eutrophication Air' and 'Eutrophication Water'")
    for eu in old_euts:
        for cf in eu.factors():
            if str(cf.flowable).lower() == 'nitrogen' and cf.context.name in ('air', 'to air'):
                if omit_n2:
                    print('Omitting gaseous nitrogen to air eutrophication\n%s' % cf)
                    for loc in cf.locations:
                        new_eut.characterize(flowable=cf.flowable, ref_quantity=cf.ref_quantity, context=cf.context,
                                             value=0.0, location=loc, origin=cf.origin)
                    continue

            for loc in cf.locations:
                new_eut.characterize(flowable=cf.flowable, ref_quantity=cf.ref_quantity, context=cf.context,
                                     value=cf[loc], location=loc, origin=cf.origin)
    return new_eut


def traci_2_biogenic_co2(traci, fg, external_ref='gwp_bio_co2', indicator='kg CO2eq incl bio'):
    """
    Duplicate the GWP method, but force the duplicate to compute biogenic CO2: see
    antelope_core.implementations.quantity.do_lcia()
    antelope_core.implementations.quantity.CO2QuantityConversion
    antelope.flow.Flow.is_co2
    :param traci:
    :param fg:
    :param external_ref:
    :param indicator:
    :return:
    """
    try:
        return fg.get(external_ref)
    except EntityNotFound:
        pass

    old_gwp = traci.get('Global Warming Air')
    new_gwp = fg.new_quantity('Global Warming Air - with biogenic CO2', ref_unit=indicator,
                              external_ref=external_ref,
                              Method=old_gwp['Method'], Category='Global Warming Air - including biogenic CO2',
                              Indicator=indicator,
                              uuid='64583b7d-bdd0-4d44-ae53-494d5b192606',
                              Comment='TRACI GWP method with biogenic CO2 enforced',
                              quell_biogenic_co2=False)
    old_gwp['quell_biogenic_co2'] = True

    for cf in old_gwp.factors():
        for loc in cf.locations:
            new_gwp.characterize(flowable=cf.flowable, ref_quantity=cf.ref_quantity, context=cf.context,
                                 value=cf[loc], location=loc, origin=cf.origin)
    return new_gwp


def traci_2_biogenic_co2_only(traci, fg, external_ref='gwp_bio_co2_only', indicator='kg CO2-bio'):
    """
    Create a new method that *only* includes biogenic CO2 flows
    :param traci:
    :param fg:
    :param external_ref:
    :param indicator:
    :return:
    """
    try:
        return fg.get(external_ref)
    except EntityNotFound:
        pass

    old_gwp = traci.get('Global Warming Air')
    bio_c_only = fg.new_quantity('Global Warming Air - biogenic CO2 Only', ref_unit=indicator,
                                 external_ref=external_ref,
                                 Method=old_gwp['Method'], Category='Global Warming Air - only biogenic CO2',
                                 Indicator=indicator,
                                 uuid='2271523e-108f-4216-a65e-dac18ce3e83f',
                                 Comment='GWP from biogenic CO2 only (no other emissions included)',
                                 quell_biogenic_co2='only')
    old_gwp['quell_biogenic_co2'] = True

    co2 = fg.flowable('124-38-9')
    for cf in old_gwp.factors():
        if fg.flowable(cf.flowable) is co2:
            for loc in cf.locations:
                bio_c_only.characterize(flowable=cf.flowable, ref_quantity=cf.ref_quantity, context=cf.context,
                                        value=cf[loc], location=loc, origin=cf.origin)
    return bio_c_only



"""
Positive-negative chart.  Draws a vertically-oriented "waterfall-lite" chart with one bar pointing up containing
all the positive segments, and another bar pointing down containing all the negative segments, so that the negative
bar ends at the net result, which is annotated.
"""

import matplotlib as mpl
import matplotlib.pyplot as plt

from pandas import DataFrame, MultiIndex


from .base import save_plot, net_color, standard_labels  # open_ylims,
from .waterfall import random_color
from antelope_core.autorange import AutoRange


mpl.rcParams['patch.force_edgecolor'] = True
mpl.rcParams['errorbar.capsize'] = 3
mpl.rcParams['grid.color'] = 'k'
mpl.rcParams['grid.linestyle'] = ':'
mpl.rcParams['grid.linewidth'] = 0.5
mpl.rcParams['lines.linewidth'] = 1.0
mpl.rcParams['axes.autolimit_mode'] = 'round_numbers'
mpl.rcParams['axes.xmargin'] = 0
mpl.rcParams['axes.ymargin'] = 0


class _PosNegAxes(object):

    def __init__(self, ax, size, qty, span, bar_width=0.28, autorange=False, color=None, fontsize=10):

        self._fontsize = fontsize

        self._ax = ax
        self._size = size
        self._qty = qty

        if not color:
            if qty.has_property('color'):
                color = qty['color']
            else:
                color = random_color(qty.uuid)

        self._color = color

        x, y = span  # to confirm it's a 2-tuple
        self._span = x, y
        self._bw = bar_width

        self._pos_handle = None
        self._neg_handle = None  # for legend

        if autorange:
            a = AutoRange(self._span[1] - self._span[0])
            self._ar_scale = a.scale
            self._unit = a.adj_unit(qty.unit)
        else:
            self._ar_scale = 1.0
            self._unit = qty.unit

        ylim = [x, y + .065 * (y - x)]  # push out the top limit
        self._ax.set_ylim([k * self._ar_scale for k in ylim])


    @property
    def _tgap(self):
        """
        Useful only for vertical bars.  We want to add 3pt, which is 1/18in.  span / size" = x / 1/18"
        :return:
        """
        return (self._span[1] - self._span[0]) * self._ar_scale / (self._size * 18)

    def draw_pos_neg(self, x, pos, neg, num_format, pos_err=0.0):

        pos *= self._ar_scale
        pos_top = pos + pos_err * self._ar_scale
        neg *= self._ar_scale

        h = self._ax.bar(x, pos, align='center', width=0.85 * self._bw, color=self._color)
        if self._pos_handle is None:
            self._pos_handle = h
        if neg != 0:
            self._ax.text(x + 0.5 * self._bw, pos_top + self._tgap, num_format % pos, ha='center', va='bottom',
                          fontsize=self._fontsize)
            x += self._bw
            h = self._ax.bar(x, neg, bottom=pos, width=0.62 * self._bw, align='center', color=net_color,
                       linewidth=0)
            if self._neg_handle is None:
                self._neg_handle = h
            # edge line
            self._ax.plot([x, x - self._bw], [pos, pos], color=(0.3, 0.3, 0.3), zorder=-1, linewidth=0.5)

            x += self._bw
            tot = pos + neg
            self._ax.plot([x, x], [0, tot], color='k', marker='_')
            self._ax.text(x, 0.5 * tot, num_format % tot, ha='center', va='center',
                          bbox=dict(boxstyle='square,pad=0.025', fc='w', ec='none'),
                          fontsize=self._fontsize)
            self._ax.plot([x, x + self._bw], [0, 0], linewidth=0)

            # edge line
            self._ax.plot([x, x - self._bw], [tot, tot], color=(0.3, 0.3, 0.3), zorder=-1, linewidth=0.5)
        else:
            self._ax.text(x, pos, '%3.2g' % pos,
                          fontsize=self._fontsize)

    def draw_error_bars(self, x, pos, pos_err, neg, neg_err):
        pos *= self._ar_scale
        neg *= self._ar_scale
        pe = (self._ar_scale * k for k in pos_err)
        ne = (self._ar_scale * k for k in neg_err)

        xs = [x, x + self._bw]
        ys = [pos, pos + neg]
        yerr = list(zip(pe, ne))
        self._ax.errorbar(xs, ys, yerr=yerr, fmt='.', color='#305040')

    def finish(self, legend=True):

        self._ax.spines['top'].set_visible(False)
        self._ax.spines['bottom'].set_visible(False)
        self._ax.spines['left'].set_visible(True)
        self._ax.spines['left'].set_linewidth(2)
        self._ax.spines['right'].set_visible(False)

        self._ax.plot(self._ax.get_xlim(), [0, 0], 'k', linewidth=2, zorder=-1)
        # open_ylims(self._ax, margin=0.05)  # this issue has supposedly been fixed?
        self._ax.set_ylabel(self._unit,
                            fontsize=self._fontsize)

        if legend and self._neg_handle is not None:
            self._ax.legend((self._pos_handle, self._neg_handle), ('Impacts', 'Avoided'))

        self._ax.set_title(self._qty['ShortName'], fontsize=self._fontsize + 2)

    @property
    def name(self):
        return self._qty['Name']

    @property
    def unit(self):
        return self._unit


class PosNegChart(object):
    """
    A PosNeg Chart draws the sum of forward and avoided burdens for each result object, together on the same
    axes, with an annotated net total.
    """

    def __init__(self, *args, horiz=False, size=4, aspect=0.4, bar_width=0.28, filename=None,
                 num_format='%3.2g', legend=True, **kwargs):
        """
        aspect reports the aspect ratio of a single chart.  aspect + bar_width together determine the aspect
        ratio of multi-arg charts.

        :param args: one or more LciaResult objects
        :param color:
        :param horiz:
        :param size:
        :param aspect:
        :param bar_width:
        :param filename:
        :param num_format:
        :param kwargs: color, autorange, fontsize...
        :param autorange:
        """
        self._pos = []
        self._neg = []
        self._idx = []

        ptr = bar_width
        for i, arg in enumerate(args):
            _pos = 0.0
            _neg = 0.0
            for c in arg.keys():
                comp = arg[c]
                val = comp.cumulative_result

                # need to check sign of fragment flow, not sign of result, if possible
                # discriminant: what determines the sign
                if hasattr(comp.entity, 'node_weight'):
                    # note: this fails on aggregated results because they don't have FragmentFlow entities
                    disc = comp.entity.node_weight
                else:
                    disc = val

                if disc > 0:
                    _pos += val
                else:
                    _neg += val

            self._pos.append(_pos)
            self._neg.append(_neg)
            self._idx.append(ptr)

            if _neg != 0:
                ptr += 2 * bar_width

            ptr += (1 - 2 * bar_width)

        span = [min(self._neg), max(self._pos)]

        cross = size * aspect * (ptr + bar_width)
        if horiz:
            fig = plt.figure(figsize=[size, cross])
        else:
            fig = plt.figure(figsize=[cross, size])

        ax = fig.add_axes([0, 0, 1.0, 1.0])

        qty = args[0].quantity

        if filename is None:
            filename = 'pos_neg_%.3s.eps' % qty.uuid

        self._pna = _PosNegAxes(ax, size, qty, span, bar_width=bar_width, **kwargs)

        for i, arg in enumerate(args):
            if horiz:
                raise NotImplementedError
                # self._pos_neg_horiz(ax, i)
            else:
                self._pna.draw_pos_neg(self._idx[i], self._pos[i], self._neg[i], num_format=num_format)

        standard_labels(ax, [arg.scenario for arg in args], ticks=self._idx, rotate=False, width=22)
        self._pna.finish(legend=legend)

        if filename != 'none':
            save_plot(filename)


class PosNegCompare(object):
    def __init__(self, *args, size=4, aspect=0.4, bar_width=0.28, filename=None,
                 num_format='%3.2g', legend=False, color=None, **kwargs):
        """
        A slightly different version, where different results are assumed to have different quantities and each
        is drawn on its own axes, but the spans of all axes are set to match the maximal pos/neg ratio (so the
        horiz axes should align)
        :param args: a sequence of LciaResult objects
        :param size: nominal plot height in inches
        :param aspect:  nominal aspect ratio of each pos/neg axis
        :param bar_width: nominal bar width
        :param filename: to save. default 'pos_neg_compare.eps'. To suppress save, pass 'none' (literal string) as filename
        :param num_format: default %3.2g
        :param csv_file: [None] if non-None, write a summary table to a csv file
        :param color: a sequence of bar color specs (RGB 3-tuples) the same length as the args.  If omitted, colors
        will be drawn first from each quantity's 'color' property, or else a random color based on the quantity's UUID
        :param kwargs: autorange, fontsize
        :param legend:
        """
        self._pos = []
        self._neg = []
        _ratios = []

        for i, arg in enumerate(args):
            _pos = 0.0
            _neg = 0.0
            for c in arg.keys():
                val = arg[c].cumulative_result
                if val > 0:
                    _pos += val
                else:
                    _neg += val

            self._pos.append(_pos)
            self._neg.append(_neg)

            if _pos + _neg < 0:
                _ratios.append( -1 * (_pos + _neg) / _pos)
            else:
                _ratios.append(0)

        print(_ratios)
        max_ratio = max(_ratios)

        n = len(args)
        cross = size * aspect * n * 1.4

        fig = plt.figure(figsize=[cross, size])

        self._pna = []

        for i, arg in enumerate(args):

            if color:
                try:
                    c = color[i]
                except (IndexError, TypeError, AttributeError):
                    print('%d failed colorspec' % i)
                    c = None
            else:
                c = None
            ax = fig.add_axes([i/n, 0, 0.8/n, 1.0])
            qty = arg.quantity

            span = (-1 * max_ratio * self._pos[i], self._pos[i])
            print(span)

            self._pna.append(_PosNegAxes(ax, size, qty, span, bar_width=bar_width, color=c, **kwargs))
            self._pna[i].draw_pos_neg(1, self._pos[i], self._neg[i], num_format=num_format)

            self._pna[i].finish(legend=legend)
            ax.set_xticks([])
            ax.set_yticks([])

        if filename is None:
            filename = 'pos_neg_compare.eps'
        if filename != 'none':
            save_plot(filename)

    @property
    def dataframe(self):
        df = DataFrame((self._table_entry(k) for k in range(len(self._pna))),
                       index=[k.name for k in self._pna])
        mc = MultiIndex.from_tuples(((' ', 'Unit'), ('Total Incurred', 'Impacts'), ('Total Avoided', 'Impacts'),
                                     (' ', 'Net Total')))
        df.columns = mc
        return df

    def _table_entry(self, index):
        """

        :param index:
        :return:
        """
        p = self._pos[index]
        n = self._neg[index]
        return {'Unit': self._pna[index].unit,
                'Total Incurred Impact': '%3.2g' % p,
                'Total Avoided Impact': '%3.2g' % n,
                'Net Total': '%3.2g' % (p+n)}


class PosNegCompareError(object):
    def __init__(self, *args, size=4, aspect=0.4, bar_width=0.28, filename=None,
                 num_format='%3.2g', legend=False, color=None, **kwargs):
        """
        Adds errorbars to above.  This should drop-in replace PosNegCompare because it is a proper superset of
        functionality with a workalike interface.

        :param args: a sequence of *3-tuples* of LciaResult objects containing main result and sensitivity results
        :param size: nominal plot height in inches
        :param aspect:  nominal aspect ratio of each pos/neg axis
        :param bar_width: nominal bar width
        :param filename: to save. default 'pos_neg_compare.eps'. To suppress save, pass 'none' (literal string) as filename
        :param num_format: default %3.2g
        :param csv_file: [None] if non-None, write a summary table to a csv file
        :param color: a sequence of bar color specs (RGB 3-tuples) the same length as the args.  If omitted, colors
        will be drawn first from each quantity's 'color' property, or else a random color based on the quantity's UUID
        :param kwargs: autorange, fontsize
        :param legend:
        """
        self._pos = []
        self._neg = []

        self._pos_err = []
        self._neg_err = []

        _ratios = []
        _overdraw = []

        def _get_pos_neg(_arg):
            """returns the total scores of positive and negative components from the LciaResult """
            _pos_ = 0.0
            _neg_ = 0.0
            for comp in _arg.keys():
                val = _arg[comp].cumulative_result
                if val > 0:
                    _pos_ += val
                else:
                    _neg_ += val
            return _pos_, _neg_

        for i, arg in enumerate(args):
            if isinstance(arg, tuple):
                _pos, _neg = _get_pos_neg(arg[0])  # pos and neg for base result
                _pos1, _neg1 = _get_pos_neg(arg[1])  # pos and neg for first sensitivity
                _pos2, _neg2 = _get_pos_neg(arg[2])  # pos and neg for second sensitivity

                _pos_err_min = max([_pos - _pos1, _pos - _pos2, 0.0])  # size of negative anomaly on positive bar
                _pos_err_max = max([_pos1 - _pos, _pos2 - _pos, 0.0])  # size of the positive anomaly on positive bar

                _neg_err_min = max([_neg - _neg1, _neg - _neg2, 0.0])  # size of negative anomaly on negative bar
                _neg_err_max = max([_neg1 - _neg, _neg2 - _neg, 0.0])  # size of the positive anomaly on negative bar

                _neg_est = min([_neg, _neg1, _neg2])  # lowest negative extent
                _pos_est = max([_pos, _pos1, _pos2])  # greatest positive extent

                self._pos_err.append((_pos_err_min, _pos_err_max))
                self._neg_err.append((_neg_err_min, _neg_err_max))

                self._pos.append(_pos)
                self._neg.append(_neg)

            else:
                res = arg

                _pos, _neg = _get_pos_neg(res)

                self._pos.append(_pos)
                self._neg.append(_neg)

                self._pos_err.append((0,0))
                self._neg_err.append((0,0))

                _pos_est = _pos
                _neg_est = _neg

            nz_pos = _pos or 1.0
            if _pos + _neg_est < 0:
                _ratios.append(-1 * (_pos + _neg_est) / nz_pos)
            else:
                _ratios.append(0)

            if 1:  # _pos_est > _pos:
                _overdraw.append(_pos_est / nz_pos)

        print(_ratios)
        print(_overdraw)
        max_ratio = max(_ratios)
        max_over = max(_overdraw)

        n = len(args)
        cross = size * aspect * n * 1.4

        fig = plt.figure(figsize=[cross, size])

        self._pna = []

        for i, arg in enumerate(args):
            if isinstance(arg, tuple):
                try:
                    qty = arg[0].quantity
                except AttributeError:
                    print(i)
                    print(arg[0])
                    raise
            else:
                qty = arg.quantity

            if color:
                try:
                    c = color[i]
                except (IndexError, TypeError, AttributeError):
                    print('%d failed colorspec' % i)
                    c = None
            else:
                c = None

            ax = fig.add_axes([i/n, 0, 0.8/n, 1.0])

            span = (-1 * max_ratio * self._pos[i], max_over * self._pos[i])
            print(span)

            self._pna.append(_PosNegAxes(ax, size, qty, span, bar_width=bar_width, color=c, **kwargs))
            self._pna[i].draw_pos_neg(1, self._pos[i], self._neg[i], num_format=num_format, pos_err=self._pos_err[i][1])

            if any(self._pos_err[i] + self._neg_err[i]):
                try:
                    self._pna[i].draw_error_bars(1, self._pos[i], self._pos_err[i], self._neg[i], self._neg_err[i])
                except ValueError:
                    print('Value error on %d:\n %g\n %g\n %g\n %g' % (i, self._pos_err[i][0], self._pos_err[i][1],
                                                                      self._neg_err[i][0], self._neg_err[i][1]))

            self._pna[i].finish(legend=legend)
            ax.set_xticks([])
            ax.set_yticks([])

        if filename is None:
            filename = 'pos_neg_compare.eps'
        if filename != 'none':
            save_plot(filename)

    @property
    def dataframe(self):
        df = DataFrame((self._table_entry(k) for k in range(len(self._pna))),
                       index=[k.name for k in self._pna])
        mc = MultiIndex.from_tuples(((' ', 'Unit'), ('Incurred', 'Impacts'), ('Avoided', 'Impacts'),
                                     (' ', 'Net Total')))
        df.columns = mc
        return df

    def _table_entry(self, index):
        """

        :param index:
        :return:
        """
        scale = self._pna[index]._ar_scale
        if scale == 1:
            num_fmt = '%3.2g'
        else:
            num_fmt = '%.3g'
        p = self._pos[index] * scale
        n = self._neg[index] * scale
        return {'Unit': self._pna[index].unit,
                'Incurred Impact': num_fmt % p,
                'Avoided Impact': num_fmt % n,
                'Net Total': num_fmt % (p+n)}



import locale
import logging

locale.setlocale(locale.LC_ALL, locale.getdefaultlocale())


def try_float(thing):
    """
    fail loudly: fallback to atof, but raise exception if that fails
    :param thing:
    :return:
    """
    try:
        value = float(thing)
    except ValueError:
        value = locale.atof(thing)
    return value


def to_float(thing):
    """
    fail silently: catch exceptions and return 0.0 if they arise
    :param thing:
    :return:
    """
    try:
        return try_float(thing)
    except ValueError:
        logging.warning('float value error "%s" - using 0.0' % thing)
        return 0.0
    except TypeError:
        logging.warning('float type error %s (%s) - using 0.0' % (thing, type(thing)))
        return 0.0



"""
Closely related to the stackbar compare, only we change our expectations for cases and units.
"""

import pandas as pd
import numpy as np
from matplotlib.cm import get_cmap
from matplotlib import pyplot as plt
from antelope_core.lcia_results import q_contrib, q_percent


def lcia_contribution_chart(df, indicator='Quantity', f_u='result', filename=None, _qs=None, stage_colors=None, colormap='viridis',
                            cmap_factor=1.0, wspace=0.3, percent=False):
    """
    "cases" is really quantities, and the only quantity is 'contrib'
    :param df:
    :param indicator: the column that contains the bar groups
    :param f_u:
    :param filename:
    :param _qs:
    :param stage_colors:
    :param colormap:
    :param cmap_factor:
    :param wspace:
    :param contrib:
    :return:
    """
    cases = []
    for i, k in df.iterrows():
        if k[indicator] not in cases:
            cases.append(k[indicator])

    # Pivot data to calculate stacked bar components
    # we want to preserve the order in the dataframe

    pivot_df = df.pivot_table(index=indicator, columns='Stage', values='Result', aggfunc='sum').fillna(
        0).reindex(cases)

    if percent:
        quantity = q_percent
        pivot_df = pivot_df.div(pivot_df.sum(axis=1) / 100, axis=0)

    else:
        quantity = q_contrib
        pivot_df = pivot_df.div(pivot_df.sum(axis=1), axis=0)

    stages = pivot_df.columns
    if stage_colors is None:
        _cm = get_cmap(colormap)
        stage_colors = {stage: _cm(i / len(stages) / cmap_factor) for i, stage in enumerate(stages)}

    # Create the figure with multiple axes
    fig, ax = plt.subplots(1, 1, figsize=(2 * len(cases), 6))
    plt.subplots_adjust(wspace=wspace)

    # Compute cumulative sums for stacking
    cumulative_sums = pivot_df.cumsum(axis=1)
    bar_positions = np.arange(len(cases))  # X positions for bars

    # Plot stacked bars
    for i, stage in enumerate(stages[::-1]):  # reverse direction for legend order
        ix = len(stages) - i - 2  # -1 for zero-indexing; -1 for cumsum offset
        bottom = cumulative_sums.iloc[:, ix] if ix >= 0 else None
        ax.bar(bar_positions, pivot_df[stage], label=stage, bottom=bottom, width=0.8,
               color=stage_colors[stage])

    # Customize the axes
    ax.set_title(f"{f_u}: {quantity.name}")
    ax.set_xlabel("Indicator")
    ax.set_ylabel(quantity.unit)
    ax.set_xlim([bar_positions[0] - 0.5, bar_positions[-1] + 0.5])
    ax.set_xticks(bar_positions)
    ax.set_xticklabels(cases, rotation=45)
    ax.legend(title="Stage", loc='upper left', bbox_to_anchor=(1, 1))

    # Set common y-label
    # fig.supylabel("Result")

    # Adjust layout
    # plt.tight_layout()
    if filename:
        plt.savefig(filename, bbox_inches='tight', dpi=300)
    return fig



from setuptools import setup

setup()

"""
Version History

0.2.3 - correct 2 bugs in QuickAndEasy workflow: keep zero tap values, cast descend to bool in model maker

0.2.2 - quick and easy add taps by recipe; node updater add targets w filter by spatial scope

0.2.1 - catch-up release for ISSST workshop

0.2.0 - catch-up release for ACLCA workshop

0.1.1post99 - 2024-05-15 get ready for PyPI release

0.1.1 - First split from lca-tools

"""



from synonym_dict import SynonymDict, TermExists


class StageManager(object):
    """
    A class to handle automatic mapping of stage names, to be used for LciaResult aggregation

    two modes: read-only and writable

    Read-only just reads in the table and executes a mapping of the input string to the selected column
    Writable mode takes a list of stage names as inputs and updates the existing mappings, then writes it to the sheet
    (if supported)
    """

    _xlsx = None
    _sheet = None

    _maps = None
    skip_values = {None, 0, 'NA'}

    @classmethod
    def from_fragment(cls, frag):
        """
        Creates (or updates) a StageManager spreadsheet based on the stages encountered in one fragment and its
        child fragments
        :param frag:
        :return:
        """
        return NotImplemented

    def __init__(self, xlsx, sheetname='stage_names', write=True, default_name='StageName', ignore_case=True):
        self._xlsx = xlsx
        self._sheetname = sheetname
        self._w = bool(write)
        if sheetname not in xlsx.sheet_names():
            if self._w:
                self._xlsx.create_sheet(sheetname)
                self._xlsx.write_row(sheetname, 0, (default_name,))
            else:
                raise KeyError('Sheet %s is missing and write is disabled' % sheetname)

        self._maps = []  # the list of column headers-- these correspond to different mappings of the same set of stages
        self._default = default_name  # the
        self._mappings = dict()
        self._sd = SynonymDict(ignore_case=ignore_case)
        self._sd._ignore_case = ignore_case

        self.read_mappings()

    def _values(self, _adict):
        return filter(lambda x: x not in self.skip_values, _adict.values())

    @property
    def keys(self):
        for k in sorted(self._mappings.keys()):
            yield k

    @property
    def map_columns(self):
        for col in self._maps[1:]:
            yield col

    def drop_column(self, col_name):
        if col_name in self._maps:
            for k in self.keys:
                self._mappings[k].pop(col_name, None)
            self._maps.remove(col_name)

    def _create_or_update_entry(self, row, default=None):
        if default is None:
            default = row[self._default]
        try:
            # update
            key = self._sd[default]
            for v in self._values(row):
                try:
                    self._sd.add_synonym(key, v)
                except TermExists:
                    print('%s: Skipping existing term "%s")' % (default, v))
            return key
        except KeyError:
            # create-- prune = True because
            ent = self._sd.new_entry(default, *self._values(row), prune=True)
            return ent.object  # a synonym set

    def read_mappings(self):
        self._sheet = self._xlsx.sheet_by_name(self._sheetname)
        self._maps = [k.value for k in self._sheet.row(0)]
        if self._default not in self._maps:
            print('warning: default name %s not present in mappings; using %s' % (self._default, self._maps[0]))
            self._default = self._maps[0]
        for i in range(1, self._sheet.nrows):
            row = self._sheet.row_dict(i)
            key = self._create_or_update_entry(row)
            self._mappings[key] = row

    def map_stage_name(self, term, column=1):
        if isinstance(column, str):
            target = column
        else:
            target = self._maps[int(column)]
        try:
            key = self._sd[term]
            mapp = self._mappings[key]
            value = mapp[target]
            return value
        except KeyError:
            return term

    def retrieve(self, lookup, col=None):
        try:
            key = self._sd[lookup]
        except KeyError:
            try:
                key = next(self._sd.objects_with_string(lookup))
            except StopIteration:
                raise KeyError(lookup)
            except TypeError:
                raise ValueError(lookup)
        if col is None:
            return self._mappings[key]
        return self.map_stage_name(key, col)

    def add_map_column(self, col):
        if col in self._maps:
            return  # nothing to do
        else:
            self._maps.append(col)

    def update_mapping(self, term, **kwargs):
        try:
            key = self._sd[term]
            d = self._mappings[key]
            d.update(kwargs)
            self._create_or_update_entry(d, default=key)
        except KeyError:
            d = {self._default: term}
            key = self._create_or_update_entry(kwargs, default=term)
            d.update(kwargs)
        for col, val in kwargs.items():
            self.add_map_column(col)
        self._mappings[key] = d

    def add_synonym(self, term, synonym):
        """

        :param term: existing term
        :param synonym: new synonym
        :return:
        """
        self._sd.add_synonym(term, synonym)

    def prune_keys(self, keys):
        s = set(keys)
        p = set()
        for k in self.keys:
            if k not in s:
                p.add(k)
        for k in p:
            print('Removing %s' % k)
            self._mappings.pop(k)

    def write_mappings(self):
        if not self._w:
            raise AttributeError('Non-writeable')

        self._xlsx.clear_region(self._sheetname)

        self._xlsx.write_row(self._sheetname, 0, self._maps)

        def _row_gen(_key):
            yield _key
            for col in self._maps[1:]:
                _map = self._mappings[_key]
                if col in _map:
                    yield _map[col]
                else:
                    yield None

        def _table_gen():
            for _k in self.keys:
                yield _row_gen(_k)

        self._xlsx.write_rectangle_by_rows(self._sheetname, _table_gen(), start_row=1)

    def __getitem__(self, item):
        return self.retrieve(item)



class LcaModelInterface(object):

    def transporters(self):
        """
        List the transporters in the model as name, link tuples
        :return: generate name, link tuples or EntitySpec namedtuples
        """
        raise NotImplementedError

    def products(self):
        """
        List the MFA products in the model as name, link tuples
        :return: generate name, link tuples or EntitySpec namedtuples
        """
        raise NotImplementedError

    @property
    def scope(self):
        raise NotImplementedError

    @property
    def ref(self):
        raise NotImplementedError

    @property
    def reference_flow(self):
        raise NotImplementedError

    @property
    def default_term_map(self):
        return dict()



"""
LcaModel class-- specifies what makes up the core model
I don't think this is used anywhere
"""

from antelope import EntityNotFound, q_node_activity
from .conventions import logistics_summary_ref, transport_model

from .interface import LcaModelInterface
from .markets import build_market_mix

from collections import namedtuple


EntitySpec = namedtuple('EntitySpec', ('link', 'ref', 'name', 'group'))


class LcaModel(LcaModelInterface):
    def __init__(self, fg, models, activity_container='Activity Container'):
        """

        :param fg: The LcForeground where the model is built
        :param models: the LcForeground where supporting models are stored
        :param activity_container: external ref for activity_container fragment
        """
        self._fg = fg
        self._models = models
        self._ac = activity_container

    @property
    def fg(self):
        return self._fg

    @property
    def models(self):
        return self._models

    # conventional names
    @property
    def _act(self):
        return q_node_activity(self._fg)

    @property
    def _fu_name(self):
        return 'MFA Activity (%s)' % self.scope

    @property
    def lca_entities(self):
        """
        Generates entity definitions suitable for fg.add_or_retrieve: link, ref, name[, group]
        where link is local, ref is the reference quantity or reference unit, name is for printing,
        group [stage] is optional
        :return: generate a list of link, ref, name, group tuples
        """

        for name, link in self.transporters():
            yield EntitySpec(link, 't*km', name, 'Transporters')

        for name, link in self.products():
            yield EntitySpec(link, 'mass', name, 'MFA Products')

    def new_functional_unit(self, name, external_ref=None, **kwargs):
        """
        This should really be a foreground method
        :param name:
        :param external_ref:
        :param kwargs:
        :return:
        """
        if external_ref is None:
            fu = self.fg.new_flow(name, self._act, **kwargs)
        else:
            fu = self.fg.add_or_retrieve(external_ref, self._act, name, strict=True, **kwargs)
        return fu

    def validate_map(self, mapping):
        """
        An MFA study is defined by maps which take MFA flows to LCA terminations.  Given a mapping, this function
        returns True if every key in the mapping is generated by lca_entities
        :param mapping:
        :return:
        """
        fail = 0
        es = [d[0] for d in self.lca_entities]
        for k in mapping.keys():
            if k not in es:
                print('Missing key: %s' % k)
                fail += 1
        return fail == 0

    @property
    def functional_unit(self):
        return self.new_functional_unit(self._fu_name, external_ref=self.ref)

    def _make_activity_container(self):
        f_u = self.functional_unit
        a_c = self.fg.new_fragment(f_u, 'Output', knobs=dict())
        self.fg.observe(a_c, name=self._ac)
        return a_c

    def _check_activity_container(self):
        return self.fg.get(self._ac)

    @property
    def activity_container(self):
        try:
            return self._check_activity_container()
        except EntityNotFound:
            return self._make_activity_container()

    @property
    def knobs(self):
        a = self.activity_container
        knobs_map = a['knobs']
        knobs = {k: (self._fg[v] or self._models[v]) for k, v in knobs_map.items()}
        return knobs

    def add_knob(self, knob, fragment):
        a = self.activity_container
        print('Adding knob %s -> %s' % (knob, fragment.name))
        a['knobs'][knob] = fragment.uuid  # we use the uuid so it will serialize-- this

    def has_knob(self, knob):
        a = self.activity_container
        return knob in a['knobs']

    def logistics_summary(self, sc_frag):
        log_ref = logistics_summary_ref(sc_frag)

        if self._fg[log_ref] is None:
            net_inflow = sum(k.value for k in sc_frag.cutoffs() if k.direction == 'Input' and k.unit == 'kg')

            f = self._fg.new_fragment(self.reference_flow, 'Input', value=net_inflow)
            self._fg.new_fragment(self.reference_flow, 'Output', balance=True, parent=f)
            for k in sc_frag.cutoffs():
                if k.direction == 'Input' and k.unit == 't*km':
                    t = self._fg.new_fragment(k.flow, k.direction, value=k.value, parent=f)
                    self._fg.observe(t)
            self._fg.observe(f, name=log_ref)

        return self._fg[log_ref]

    def mfa_freight_market(self, flow, recipe):
        frag = transport_model(flow)
        if self._fg[frag] is None:
            print('Making new freight market fragment %s' % flow.external_ref)
            k = self._fg.new_fragment(flow, 'Output')
            if build_market_mix(self._fg, k, self._models, recipe, freight=True):
                self._fg.observe(k, name=frag)
            else:
                print('FAILED MARKET MIX: Not building model %s' % frag)
                self._fg.delete_fragment(k)
        return self._fg[frag]




import pandas as pd

from .components_mixin import ComponentsMixin
from .lca_model_runner import LcaModelRunner

from antelope_foreground.fragment_flows import group_ios, ios_exchanges, frag_flow_lcia

from functools import reduce


class ScenarioRunner(ComponentsMixin, LcaModelRunner):
    """
    This runs a single model (fragment), applying a set of different scenario specifications. 
    """
    def _scenario_tuple(self, arg):
        """
        Translates None, strings, or tuples into tuples. should return a flat (non-nested) tuple
        check out all the work being done by that one comma
        :param arg:
        :return:
        """
        if arg:
            if isinstance(arg, tuple):
                return tuple(reduce(lambda a, b: a+b, (self._scenario_tuple(k) for k in arg)))
            else:
                return arg,
        return ()

    def __init__(self, model, *common_scenarios, **kwargs):
        """

        :param kwargs: agg_key, default is lambda x: x['StageName']
        """
        super(ScenarioRunner, self).__init__(**kwargs)

        self._common_scenarios = set()

        self._model = model
        self._traversals = dict()

        self._params = dict()

        for scenario in common_scenarios:
            self.add_common_scenario(scenario)

    @property
    def model(self):
        return self._model

    def params(self, scenario):
        return self._params[scenario]

    def add_common_scenario(self, scenario):
        if isinstance(scenario, tuple):
            for sc in scenario:
                self._common_scenarios.add(sc)
        else:
            self._common_scenarios.add(scenario)
        self.recalculate()

    def remove_common_scenario(self, scenario):
        self._common_scenarios.remove(scenario)
        self.recalculate()

    def traverse_all(self):
        for case in self.scenarios:
            self._traverse_case(case)

    def _traverse_case(self, case):
        print('traversing %s' % case)
        sc = self._params[case]
        sc_apply = sc + tuple(self.common_scenarios)
        self._traversals[case] = list(self._model.traverse(sc_apply))

    def _recalculate_case(self, case, **kwargs):
        self._traverse_case(case)
        for q in self.lcia_methods:
            self.run_lcia_case_method(case, q, **kwargs)
        for w in self.weightings:
            self._run_case_weighting(case, w)

    def add_case_param(self, case, param):
        if case not in self._params:
            raise KeyError('Unknown case %s' % case)
        assert isinstance(param, str)
        if param in self._params[case]:
            print('Param %s already in case %s' % (param, case))
        else:
            self._params[case] += (param, )
        self._recalculate_case(case)

    def remove_case_param(self, case, param):
        if case not in self._params:
            raise KeyError('Unknown case %s' % case)
        assert isinstance(param, str)
        if param not in self._params[case]:
            raise ValueError('Param %s not in case %s' % (param, case))
        else:
            self._params[case] = tuple(filter(lambda x: x != param, self._params[case]))
        self._recalculate_case(case)

    @property
    def common_scenarios(self):
        for k in sorted(self._common_scenarios):
            yield k

    def add_case(self, case, *params):
        self.add_scenario(case)  # raises KeyError
        self._params[case] = self._scenario_tuple(params)
        self._recalculate_case(case)

    def fragment_flows(self, scenario):
        return self._traversals[scenario]

    def cutoffs(self, scenario, **kwargs):
        ios, _ = group_ios(self._model, self._traversals[scenario], **kwargs)
        return ios_exchanges(ios, ref=self._model)

    def cutoffs_dataframe(self, include_activity=True):
        def _cutoffs_row(k, sc):
            return {'Case': sc, 'Cutoff': k.flow.name, 'Direction': k.direction, 'Magnitude': k.value,
                    'Unit': k.unit}
        if include_activity:
            return pd.DataFrame([_cutoffs_row(k, s) for s in self.scenarios for k in self.cutoffs(s)])
        else:
            return pd.DataFrame([_cutoffs_row(k, s) for s in self.scenarios for k in self.cutoffs(s)
                                 if k.unit != 'activity'])

    def activity(self, scenario):
        """
        should think about some more generalized 'activity' query, a la antelope_foreground.terminal_nodes
        :param scenario:
        :return:
        """
        return [f for f in self._traversals[scenario] if f.fragment.top() is self._model]

    def _run_scenario_lcia(self, scenario, lcia, **kwargs):
        sc = self._params[scenario]
        sc_apply = sc + tuple(self.common_scenarios)
        return frag_flow_lcia(self._traversals[scenario], lcia, scenario=sc_apply, **kwargs)





from antelope import EntityNotFound, UnknownOrigin, MultipleReferences, ConversionError, NoReference
from antelope.interfaces.iindex import InvalidDirection, comp_dir, check_direction
from antelope_foreground.terminations import FlowConversionError

from .quick_and_easy import QuickAndEasy, AmbiguousResult
from .float_conv import to_float, try_float
from .headers import PRODUCTION_HEADER
import logging
from collections import defaultdict


class ConsistencyError(Exception):
    pass


class NoInformation(Exception):
    """
    Not even enough info to make a cutoff flow
    """
    pass


class FailedTermination(Exception):
    pass


class BadExchangeValue(Exception):
    pass


class BadDisplacementRelation(Exception):
    """
    Something is wrong with the displacement table entry
    """
    pass


class DispositionError(Exception):
    pass


class ModelMaker(QuickAndEasy):
    """
    This class enhances QuickAndEasy with some useful tools for building product models:

     - create_or_retrieve_reference() - a utility that retrieves or else creates a fragment with a specified flow
     - make_production_row() - construct a single child flow from an entry in the production spreadsheet (see spec below)
     - make_production() - make/update all production, reference flows + child flows.  idempotent.
     - make_displacement_model() - make/update one model from an entry in the displacement spreadsheet (see spec below)
     - make_displacement() - make/update all displacement models

    Production Sheet: (default 'production') Columns:
    reference spec: should be the same for every entry for a given reference flow
    (these are used in create_or_retrieve_reference())
    prod_flow	-- the external_ref of the flow that is used as the reference flow. Only one production model per flow.
    ref_direction	-- direction of the reference w.r.t the fragment (i.e. comp_dir w.r.t. use)
    ref_value	-- observed exchange value of the reference flow
    ref_unit	-- unit of measure of the observation

    child flow spec: each nonempty row generates one child flow
    (these are used in make_production_row())
    direction	-- direction of the child flow w.r.t. the parent node
    amount	-- observed exchange value of the reference flow
    balance_yn	-- boolean (could begin to accept 'balance' as a direction spec instead)
    amount_hi	-- observed exchange value for high sensitivity case
    amount_lo	-- observed exchange value for high sensitivity case
    units	-- unit of measure for amount, amount_hi, amount_lo
    child_flow  -- external ref of child flow (if blank, will be taken from anchor)
    stage_name	-- fragment 'StageName' property
    note	-- fragment 'note' property
    Comment	-- fragment 'Comment' property

    scenario	-- deprecated

    anchor spec:
    (these are used in _find_term_info(), ultimately passed to find_background_rx()
    target_origin	-- origin for anchor node
    compartment	-- used for specifying elementary flows (leave origin blank)
    target_flow -- passed to find_background_rx as flow_name_or_ref (if blank, use child_flow if specified)
    locale	-- used as SpatialScope argument in find_background_rx
    target_name	-- used as process_name argument in find_background_rx
    target_ref	-- passed to find_background_rx as external_ref

    The regression is as follows:
      if origin is specified: row describes an anchored flow
         origin == 'here' -> look in local foreground
         pass a dict to find_background_rx containing:
         target_ref (as external_ref), target_name (as process_name), flow_name_or_ref, locale (as SpatialScope)
      else:
         row describes either a cutoff or an exterior (context / environment) flow
         flow_name_or_ref is retrieved using get_local(). no other columns are used other than compartment (maybe locale)
         if compartment is specified:
            retrieve context
         else:
            create cutoff

    Displacement sheet: (default 'displacement') Columns:
    in_use	-- (bool) whether to make an entry for this row

    Defining the displacement relationship
    md_flow	-- materially-derived flow that is driving the displacement
    mdflow_refprop	-- a vlookup provided for the xlsx editor's convenience: =vlookup(md_flow,flow_ref_qty,2,false)
    refunit	-- unit of measure to define reference end of displacement relationship
    dp_flow	-- displaced-product flow that is getting driven
    dp_refprop	-- a vlookup provided for the xlsx editor's convenience: =vlookup(dp_flow,flow_ref_qty,2,false)
    dp_refunit	-- unit of measure to define the displaced end of the displacement relationship
    note	-- fragment 'note' property

    defining the displacement rate
    scenario	-- optional scenario name for
    disp_lo	-- observed displacement alpha rate for 'low-displacement' scenario
    disp_rate	-- observed displacement alpha, named (or default) scenario
    disp_hi	-- observed displacemnent alpha rate for 'high-displacement' scenario
    value	-- observed displacement beta for named (or default) scenario
    time_equiv	-- multiplier for time equivalency (not yet implemented)

    defining forward and displaced transport
    md_truck	-- truck transport for massive md flows, using truck transport target specified by trans_truck
    dp_truck	-- truck transport for massive dp flows, using truck transport target specified by trans_truck
    dp_ocean	-- ocean transport for massive dp flows, using truck transport target specified by trans_ocean

    """
    def __init__(self, *args, sens_lo='sens_lo', sens_hi='sens_hi', **kwargs):
        super(ModelMaker, self).__init__(*args, **kwargs)

        self.sens_lo = sens_lo
        self.sens_hi = sens_hi

        self._skips = []
        self._errors = dict()

    @property
    def skips(self):
        return list(self._skips)

    @property
    def errors(self):
        d = defaultdict(list)
        for k, e in self._errors.items():
            d[e.__class__.__name__].append(k)
        return dict(d)

    def error(self, k):
        return self._errors[k]

    def autodetect_flows(self, sheetname, external_ref=None, ref_quantity=None, ref_unit=None, name=None,
                         context=None, entity_uuid=None, **kwargs):
        """
        Load the designated sheet. iterate through rows, using the arguments as column mappings.

        :param sheetname: name of the source sheet
        :param external_ref:
        :param ref_quantity:
        :param ref_unit:
        :param name:
        :param context:
        :param entity_uuid:
        :param kwargs:
        :return:
        """
        if not self._quiet:
            print('\n autodetecting %s' % sheetname)

        sheet = self.xlsx[sheetname]

        if external_ref is None:
            external_ref = 'external_ref'
        if ref_quantity is None:
            if ref_unit is None:
                ref_quantity = 'referenceQuantity'
            else:
                pass  # use ref_unit
        if name is None:
            name = 'name'

        # all others are optional
        if context:
            kwargs['context'] = context
        if entity_uuid:
            kwargs['entity_uuid'] = entity_uuid

        count = 0
        new = set()

        for r in range(1, sheet.nrows):
            ssr = r + 1
            row = sheet.row_dict(r)
            ext_ref = row.get(external_ref)
            the_name = row.get(name, ext_ref)
            if ext_ref is None:
                continue
            if the_name is None:
                the_name = ext_ref
            if ref_quantity is None:
                ref_q = self._unit_map.get(row.get(ref_unit))
                if ref_q is None:
                    print('%s:%d Unrecognized unit %s- skipping flow %s' % (sheetname, ssr, row.get(ref_unit), ext_ref))
                    continue
            else:
                ref_q = row.get(ref_quantity)
            count += 1

            args = {k: row.get(v) for k, v in kwargs.items() if v is not None and row.get(v) is not None}
            if self.fg[ext_ref] is None:
                new.add(ext_ref)
            f = self.fg.add_or_retrieve(ext_ref, ref_q, the_name, **args)
            # update the name if it contains new information-- important for LCIA
            if the_name != f['name'] and the_name != ext_ref:
                f['name'] = the_name
                f.clear_chars()

        print('Reviewed %d flows (%d new added)' % (count, len(new)))

    def detect_spanner_flows(self, spanner_ref, **kwargs):
        self.autodetect_flows(spanner_ref, external_ref='flow', name='Name', ref_unit='units',
                              context='context', **kwargs)

    def detect_production_flows(self, production_sheet='production', **kwargs):
        self.autodetect_flows(production_sheet, external_ref='prod_flow', ref_unit='ref_unit', **kwargs)

    def update_flows(self, sheetname='flows'):
        """
        here we are using the same flows sheet spec as XlsxArchiveUpdater so we don't need to configure column mappings
        :param sheetname:
        :return:
        """
        sheet = self.xlsx[sheetname]
        columns = ('external_ref', 'referenceQuantity', 'Name', 'Comment', 'Compartment')
        write = [list(columns)]
        exis = set()
        error = 0

        def _make_a_row(_f):
            cx = _f.context.name
            if cx == 'None':
                cx = None
            return [_f.external_ref, _f.reference_entity['Name'], _f['Name'], _f['Comment'], cx]

        # first, existing
        for i in range(1, sheet.nrows):
            row = [k.value for k in sheet.row(i)]
            ssr = i + 1
            if len(row) == 0 or row[0] is None:
                write.append([None, None, None, None, None])
                continue
            try:
                f = self.fg.get(row[0])
                write.append(_make_a_row(f))
                exis.add(f.external_ref)
            except KeyError:
                logging.warning('unrecognized flow in existing pass (%s row %d %s)' % (sheetname, ssr, row[0]))
                write.append(row[:5])
            except TypeError:
                logging.warning('TypeError problem in existing pass (%s row %d %s)' % (sheetname, ssr, row[0]))
                error += 1

        # add new
        added = 0
        for f in self.fg.flows():
            if f.origin != self.fg.origin:
                continue
            if f.external_ref in exis:
                continue
            try:
                write.append(_make_a_row(f))
                added += 1
            except TypeError:
                logging.warning('TypeError problem - skipping new flow %s' % f.external_ref)
                error += 1
        msg = 'Writing %d existing and %d new flows to sheet %s' % (len(exis), added, sheetname)
        if error:
            msg += ' (%d errors)' % error
        print(msg)
        self.xlsx.write_rectangle_by_rows(sheetname, write)

    # GENERIC
    def _get_one(self, hits, strict=False, prefix=None):
        """
        Overrides the standard _get_one with a feature to only search among activities whose name matches a prefix-
        :param hits:
        :param strict:
        :param prefix:
        :return:
        """
        if prefix:
            f_hits = filter(lambda x: x.external_ref.startswith(prefix), hits)
            return super(ModelMaker, self)._get_one(f_hits, strict=strict)
        else:
            return super(ModelMaker, self)._get_one(hits, strict=strict)

    def create_or_retrieve_reference(self, flow_or_ref, direction='Output', external_ref=None, prefix=None):
        """
        All these functions are written to deal with poorly-specified corner cases. it's terrible.
        what do we want to do?
         - look for a fragment with the specified flow
        :param flow_or_ref:
        :param direction:
        :param external_ref:
        :param prefix:
        :return:
        """
        if hasattr(flow_or_ref, 'entity_type'):
            if flow_or_ref.entity_type == 'flow':
                flow = flow_or_ref
                frag = self._get_one(self.fg.fragments_with_flow(flow), strict=True, prefix=prefix)
            elif flow_or_ref.entity_type == 'fragment':
                frag = flow_or_ref.top()
            else:
                raise TypeError(flow_or_ref)
        else:
            flow = self.fg[flow_or_ref]
            if flow is None:
                flow = self.fg.get(flow_or_ref)  # raises EntityNotFound eventually
                # yes, this is (regrettably) opposite the convention where .get() returns None
                # raise EntityNotFound(flow_or_ref)

            try:
                frag = self._get_one(self.fg.fragments_with_flow(flow), strict=True, prefix=prefix)
            except EntityNotFound:
                if prefix:
                    external_ref = external_ref or '%s_%s' % (prefix, flow.external_ref)
                return self._new_reference_fragment(flow, direction, external_ref)

        if external_ref or prefix:
            # we want to name the fragment
            name = external_ref or '%s_%s' % (prefix, frag.flow.external_ref)
            if frag.external_ref == frag.uuid:  # not named yet
                self.fg.observe(frag, name=name)
            elif frag.external_ref != name:
                print('Warning, fragment already named %s' % frag.external_ref)
        return frag

    def _find_reference_exchange(self, row):
        # first, find termination
        org = row.get('target_origin') or row.get('origin')
        if org:
            if org == 'here':
                origin = self.fg.origin
            else:
                origin = org

            d = {'external_ref': row.get('target_ref') or row.get('external_ref'),
                 'process_name': row.get('target_name'),
                 'flow_name_or_ref': row.get('target_flow') or row.get('term_flow') or row.get('flow_name') or row.get('child_flow'),
                 'SpatialScope': row.get('locale')}  # default to RoW

            try:
                rx = self.find_background_rx(origin, **d)
            except AmbiguousResult:
                if d['SpatialScope'] is None:
                    d['SpatialScope'] = 'RoW'
                    rx = self.find_background_rx(origin, **d)
                else:
                    if not d['SpatialScope'].startswith('^'):
                        d['SpatialScope'] = '^%s$' % d['SpatialScope']
                        try:
                            rx = self.find_background_rx(origin, **d)
                        except AmbiguousResult:
                            raise AmbiguousResult(*d.values())
                    else:
                        raise AmbiguousResult(*d.values())
            except (KeyError, EntityNotFound):
                raise FailedTermination(*d.values())
        else:
            if row.get('compartment'):
                # context
                rx = self.fg.get_context(row['compartment'])
            else:
                # cutoff
                rx = None

        return rx

    def _build_production_row(self, parent, row):
        """
        I probably should break this down a little better--- so many precedence rules + heuristics
        basically, we want to do the following:
         - terminate to a foreground process: specify 'here' origin and either flow_name or external_ref
         - terminate to a cutoff: specify no origin and flow_name

        :param parent: parent fragment
        :param row: row_dict
        :return:
        """

        rx = self._find_reference_exchange(row)

        if row.get('child_flow'):
            child_flow = self.fg.get(row.get('child_flow'))
        else:
            if hasattr(rx, 'flow'):
                child_flow = rx.flow
            else:
                cf_ref = row.get('flow_name') or row.get('term_flow')
                if cf_ref:
                    child_flow = self.fg.get(cf_ref)
                else:
                    raise NoInformation

        try:
            child_direction = check_direction(row['direction'])
        except InvalidDirection:
            if hasattr(rx, 'direction'):
                child_direction = comp_dir(rx.direction)
            elif hasattr(rx, 'sense') and rx.sense is not None:
                child_direction = comp_dir(rx.sense)
            else:
                child_direction = parent.direction

        if row['balance_yn']:  # or child_direction == 'balance'
            balance = True
        else:
            balance = False

        try:
            c = next(parent.children_with_flow(child_flow))
            if balance:
                if parent.balance_flow:
                    if parent.balance_flow is not c:
                        raise ConsistencyError
                else:
                    c.set_balance_flow()
            else:
                if parent.balance_flow is c:
                    c.unset_balance_flow()
        except StopIteration:
            if balance:
                if parent.balance_flow:
                    parent.balance_flow.unset_balance_flow()
            c = self.fg.new_fragment(child_flow, child_direction, parent=parent, balance=balance)

        if not balance:
            try:
                ev = to_float(row['amount'])
            except (TypeError, ValueError):
                raise BadExchangeValue(row.get('amount'))
            try:
                self.fg.observe(c, exchange_value=ev, units=row['units'])
            except ConversionError:
                raise BadExchangeValue(c.flow.reference_entity, row['units'])
            if row.get('amount_lo', None) is not None:
                ev_lo = to_float(row['amount_lo'])
                self.fg.observe(c, exchange_value=ev_lo, units=row['units'], scenario=self.sens_lo)
            if row.get('amount_lo', None) is not None:
                ev_hi = to_float(row['amount_hi'])
                self.fg.observe(c, exchange_value=ev_hi, units=row['units'], scenario=self.sens_hi)
        """
        Logic here:
        if descend is specified, set it
        else: 
        if no stage name is specified, descend should be True
        """
        descend = row.get('descend')
        if row.get('stage_name'):
            c['StageName'] = row['stage_name']
            if descend is None:
                descend = False
            else:
                descend = bool(descend)
        else:
            if descend is None:
                descend = True
            else:
                descend = bool(descend)
        c.terminate(rx, descend=descend)

        if row.get('note'):
            c['note'] = row['note']
        if row.get('Comment'):
            c['Comment'] = row['Comment']

        tap_recipe = row.get('add_taps', None)
        if tap_recipe:
            self.apply_tap_recipes(c, tap_recipe)

        # add residual information to child flow
        for k in row.keys():
            if k not in PRODUCTION_HEADER:
                c[k] = row[k]

        return c

    def _log_e(self, ssr, e):
        if ssr in self._errors:
            logging.warning('double error %s' % ssr)
        else:
            self._errors[ssr] = e

    def _make_production_references(self, sheet, prefix):
        refs = set()
        for r in range(1, sheet.nrows):
            ssr = r + 1
            # ASSUMPTION: prod_flow is first column
            # CONVENTION: production processes are all outputs
            row = sheet.row_dict(r)
            if row.get('prod_flow'):
                dirn = row.get('ref_direction', 'Output') or 'Output'
                try:
                    ref = self.create_or_retrieve_reference(row['prod_flow'], dirn, prefix=prefix)
                    refs.add(row['prod_flow'])
                except EntityNotFound as e:
                    print('%d: unrecognized reference flow %s' % (ssr, e.args))
                    self._log_e(ssr, e)
                    continue
                try:
                    rv = try_float(row['ref_value'])
                except KeyError:
                    print('%d: skipping omitted ref_value' % ssr)
                    self._skips.append(ssr)
                    continue
                except (TypeError, ValueError):
                    print('%d: skipping bad ref_value %s' % (ssr, row['ref_value']))
                    self._skips.append(ssr)
                    continue
                ru = row.get('ref_unit')
                try:
                    self.fg.observe(ref, exchange_value=rv, units=ru)
                except ConversionError:
                    print('%d: Skipping bad unit conversion specification %s [%s]' % (ssr, ru,
                                                                                      ref.flow.reference_entity))
        return len(refs)

    def _make_production_childflows(self, sheet, prefix=None):
        count = 0
        for r in range(1, sheet.nrows):
            ssr = r + 1
            row = sheet.row_dict(r)
            if row.get('prod_flow'):
                try:
                    parent = self.create_or_retrieve_reference(row['prod_flow'], prefix=prefix)
                    c = self._build_production_row(parent, row)
                    c['_%s_row' % (prefix or sheet.name)] = ssr
                    print('== %03d ==: %s' % (ssr, c))
                    count += 1
                except NoInformation:
                    self._skips.append(ssr)
                    print('## %03d ##: No information for cutoff' % ssr)
                except FailedTermination as e:
                    self._log_e(ssr, e)
                    print('## %03d ##: Failed Termination %s' % (ssr, e.args))
                except AmbiguousResult as e:
                    self._log_e(ssr, e)
                    print('## %03d ##: Ambiguous Result %s' % (ssr, e.args))
                except UnknownOrigin as e:
                    self._log_e(ssr, e)
                    print('## %03d ##: Unknown Origin %s' % (ssr, e.args))
                except MultipleReferences as e:
                    self._log_e(ssr, e)
                    print('## %03d ##: Multiple References %s' % (ssr, e.args))
                except NoReference as e:
                    self._log_e(ssr, e)
                    print('## %03d ##: No Reference %s' % (ssr, e.args))
                except BadExchangeValue as e:
                    self._log_e(ssr, e)
                    print('## %03d ##: Bad Exchange Value %s' % (ssr, e.args))
                except EntityNotFound as e:
                    self._log_e(ssr, e)
                    print('## %03d ##: wayward entity-not-found error %s' % (ssr, e.args))
                except FlowConversionError as e:
                    self._log_e(ssr, e)
                    print('## %03d ##: flow-conversion termination error %s' % (ssr, e.args))
        return count

    def make_production(self, sheetname='production', prefix='prod', taps=None, detect_flows=True):
        """
        Strategy here:

         - production sheet includes *all* meso-scale processes
         - each production flow is provided by a single, distinct production process
         - each production record indicates a child flow with the observed amount (or balance if balance is checked)
         - terminates to the designated target (decode origin)
        :param self:
        :param sheetname: default 'production'
        :param prefix: prepend to flow_ref to get frag_ref
        :param taps: [None] if present, load taps from, named sheet after creating references but before child flows
        :return:
        """
        if self.xlsx is None:
            raise AttributeError('Please attach Google Sheet')
        self._errors = dict()  # reset errors
        if detect_flows:
            self.detect_production_flows(sheetname)

        sheet = self.xlsx[sheetname]

        # first pass: create all production fragments
        refs = self._make_production_references(sheet, prefix)

        if taps:
            self.load_taps_from_spreadsheet(taps)

        # second pass: create child flows
        count = self._make_production_childflows(sheet, prefix)
        print('Created %d reference flows with %d child flows (%d errors)' % (refs, count, len(self._errors)))

    def make_production_row(self, ssr, sheetname='production', prefix='prod'):
        """
        
        :param ssr: spreadsheet row 
        :param sheetname: 
        :param prefix: 
        :return: 
        """
        if self.xlsx is None:
            raise AttributeError('Please attach Google Sheet')

        sheet = self.xlsx[sheetname]
        row = sheet.row_dict(ssr - 1)
        if not row.get('prod_flow'):
            raise ValueError('Empty prod_flow')

        # update parent
        dirn = row.get('ref_direction', 'Output') or 'Output'
        parent = self.create_or_retrieve_reference(row['prod_flow'], direction=dirn, prefix=prefix)
        rv = try_float(row['ref_value'])
        ru = row.get('ref_unit')
        self.fg.observe(parent, exchange_value=rv, units=ru)

        # build child
        c = self._build_production_row(parent, row)
        print('== %03d ==: %s' % (ssr, c))
        c['_%s_row' % (prefix or sheet.name)] = ssr
        if ssr in self._errors:
            print('removing error for row %d' % ssr)
            self._errors.pop(ssr)
        return c

    def _check_alpha_beta_prod(self, node, row_dict):
        """
        This is meant to be a generic utility that uses several boilerplate entries in a rowdict (hmm, these could
        actually be kwargs) to construct a displacement relationship between two items.  The unit conversions happen
        externally- either upstream in the "node" when it is created, or downstream when the displaced product is
        terminated to a production activity.

        Note:
        "alpha" = economic displacement = "epsilon" in MRC
        "beta" = technical displacement = "tau" in MRC

        It should create the fragments if they don't exist; re-observe them if they do

        kwargs: td_flow, dp_flow, dp_refunit, disp_lo, disp_rate, disp_hi, value, scenario
        :param node:
        :param row_dict:
        :return:
        """
        # disp = self.fg['displacement']  # ugggg this should really be raising a key error
        # if disp is None:
        disp = self.fg.add_or_retrieve('displacement', 'Number of items', 'Displacement Rate',
                                       comment="dimensionless value used in displacement calculation",
                                       group="modeling")
        td = row_dict['md_flow']
        dp = row_dict['dp_flow']
        scenario = row_dict.get('scenario')
        alpha_name = 'epsilon-%s-%s' % (td, dp)
        beta_name = 'tau-%s-%s' % (td, dp)

        try:
            alpha = next(node.children_with_flow(disp))
        except StopIteration:
            alpha = self.fg.new_fragment(disp, 'Output', parent=node,
                                         name='Market Displacement rate (alpha)')

        """
        OK, so now we are going to define the default displacement as 0.75 * alpha
        and create high-and-low displacement of alpha and 0.5 * alpha respectively
        """
        self.fg.observe(alpha, name=alpha_name)

        a_lo = row_dict.get('disp_lo')
        a = row_dict['disp_rate']
        a_hi = row_dict.get('disp_hi')

        self.fg.observe(alpha, a, scenario=scenario)  # unitless
        if a_hi:
            self.fg.observe(alpha, a_hi, scenario='high-displacement')  # unitless
        if a_lo:
            self.fg.observe(alpha, a_lo, scenario='low-displacement')  # unitless

        try:
            beta = next(alpha.children_with_flow(disp))
        except StopIteration:
            beta = self.fg.new_fragment(disp, 'Output', parent=alpha, name='Displacement relation')

        self.fg.observe(beta, row_dict['value'], name=beta_name)

        prod = self.fg.get(dp)
        try:
            output = next(beta.children_with_flow(prod))
        except StopIteration:
            # we do this in case the dp flow has changed
            cfs = list(beta.child_flows)
            for cf in cfs:
                self.fg.delete_fragment(cf)
            output = self.fg.new_fragment(prod, 'Output', parent=beta, name='Displaced %s' % prod.name)

        self.fg.observe(output, 1.0, units=row_dict['dp_refunit'])
        return output

    def _check_transport_link(self, node, target, distance_km, scenario=None, stage_name='Transport'):
        """
        Builds or updates a transport link child of the named node (measured in mass).  Assumes units are kg;
        freight is calculated as distance_km / 1000.0, times a correction factor of the node's exchange value IF the
        node is a reference node
        :param node:
        :param target:
        :param distance_km:
        :return:
        """
        mass = self.fg.get_canonical('mass')
        if node.flow.reference_entity is not mass:
            raise DispositionError('non-mass transport flow for %s' % node)

        if target is None:
            return  # nothing to do

        if node.is_reference:
            ev = distance_km * node.observed_ev / 1000.0
        else:
            ev = distance_km / 1000.0

        try:
            cf = next(node.children_with_flow(target.flow))
        except StopIteration:
            cf = self.fg.new_fragment(target.flow, target.direction, parent=node)
            cf.terminate(target)

        self.fg.observe(cf, exchange_value=ev, scenario=scenario)
        cf['StageName'] = stage_name
        cf.term.descend = False
        return cf

    def make_displacement_model(self, row, trans_truck=None, trans_ocean=None, embed_disp_prod=True):
        """
        This creates or updates a displacement model that maps a particular product flow to a particular displaced
        flow, through an alpha-beta run, with added transport.  The alpha-beta run is standard; the other parts
        are not yet.

        :param row:
        :param trans_truck: external ref of truck transport process
        :param trans_ocean: external ref of ocean transport process
        :param embed_disp_prod: [True] whether to auto-anchor the displaced product flow if a suitable production
         activity is found.  Set to False to leave this as a cut-off flow
        :return:
        """
        product_ref = row.get('md_flow')
        product_refunit = row.get('refunit')
        disp_ref = row.get('dp_flow')

        ext_ref = 'displacement-%s-%s' % (product_ref, disp_ref)

        mass = self.fg.get_canonical('mass')
        truck_mdl = self.fg[trans_truck]
        ocean_mdl = self.fg[trans_ocean]

        td = self.fg.get(product_ref)
        dp = self.fg.get(disp_ref)

        name = 'Displacement, %s displ. %s' % (td.name, dp.name)

        # first, construct or retrieve the reference fragment
        node = self.fg[ext_ref]
        if node is None:
            node = self.fg.new_fragment(td, 'Input', Name=name, external_ref=ext_ref, StageName='Disposition')
        else:
            node.flow = td  # just to ensure
        self.fg.observe(node, exchange_value=1.0, units=product_refunit)
        node['note'] = row.get('note')

        if row.get('md_truck'):
            if td.reference_entity is mass:
                # then we can do transport--- only doing transport for massive flows
                if truck_mdl:
                    cf = self._check_transport_link(node, truck_mdl, to_float(row['md_truck']),
                                                    stage_name='Transport, %s' % td.name)
                    cf['stage_name'] = 'Transport, Products'  # stage_name kwarg becomes StageName
                else:
                    print('No truck transport model specified/found; skipping forward transport')

        output = self._check_alpha_beta_prod(node, row)

        disp = self.fg['prod_%s' % disp_ref]
        if disp:
            if embed_disp_prod:
                output.terminate(disp)

            if row.get('dp_truck'):
                if truck_mdl:
                    self._check_transport_link(disp, truck_mdl, to_float(row['dp_truck']),
                                               stage_name='Transport, Displaced')
                else:
                    print('No truck transport model specified/found; skipping displaced truck transport')

            if row.get('dp_ocean'):
                if ocean_mdl:
                    self._check_transport_link(disp, ocean_mdl, to_float(row['dp_ocean']),
                                               stage_name='Transport, Displaced')
                else:
                    print('No ocean transport model specified/found; skipping displaced ocean transport')

        return node

    def make_displacement(self, sheetname='displacement',
                          trans_truck='prod_transport_generic',
                          trans_ocean='prod_transport_ocean',
                          embed_disp_prod=True):
        """
        Here we want to replicate what we did for CATRA, only improve it.  We have a table in the spreadsheet, and
        we want to construct a disposition model for each record that is marked "in use".  That model should:
         - take the designated flow IN
         - attach the designated freight (must ensure to account for tonnes
         - attach alpha
         - attach OUT flow
         - attach

        :param sheetname: default 'displacement'
        :param trans_truck: default 'prod_transport_generic'
        :param trans_ocean: default 'prod_transport_ocean'
        :param embed_disp_prod: [True] whether to auto-anchor the displaced product flow if a suitable production
         activity is found.  Set to False to leave this as a cut-off flow
        :return:
        """
        disp = self.xlsx[sheetname]
        for r in range(1, disp.nrows):
            row = disp.row_dict(r)
            if row.get('in_use'):
                self.make_displacement_model(row, trans_truck=trans_truck, trans_ocean=trans_ocean,
                                             embed_disp_prod=embed_disp_prod)



"""
Antelope Observer
worksheets and column headings

An observer xlsx-like should have the following sheets:

<archive contents>
flows, quantities, flowproperties

<quick-and-easy>
spanners - list of annotated models
a sheet for each model named
note: all columns that appear in a model listing will be applied. add fields with abandon.

taps - flows to be tapped, and where they should be tapped to
observations - exchange values and anchors to be applied to the model

<model maker>
production - complete fragment specification

The list of headers for each sheet are provided below in a canonical order and capitalization. 
Order is not binding (row dictionary index required) but case should be observed.  
"""
from pydantic import BaseModel
import re


valid_ref = re.compile('[A-Za-z0-9-_.]+')


class SpannerMeta(BaseModel):
    """
    metadata for fragment models
    """
    '''
    apparently there's no way to sync this pydantic model with SPANNERS_META
    '''
    external_ref: str
    name: str
    description: str
    author: str
    source: str
    version: str

    @classmethod
    def from_form(cls, external_ref, name=None, description='SpannerMeta', author='observer', source='observed',
                  version='0.1'):
        assert bool(valid_ref.match(external_ref)), 'Invalid external reference %s' % external_ref

        if name is None:
            name = external_ref

        return cls(external_ref=external_ref, name=name, description=str(description), author=str(author),
                   source=str(source), version=str(version))


QUANTITIES_HEADER = ('external_ref', 'referenceUnit', 'Name', 'Comment', 'Synonyms')
FLOWS_HEADER = ('external_ref', 'referenceQuantity', 'Name', 'Comment', 'Compartment')
FLOWPROPERTIES_HEADER = ('flow', 'ref_quantity', 'ref_unit', 'quantity', 'unit', 'value', 'source', 'note')

SPANNERS_META = ('external_ref', 'name', 'description', 'author', 'source', 'version')

SPANNER_HEADER = ('flow', 'direction', 'Name', 'amount', 'units', 'amount_hi', 'amount_lo', 'context', 'descend',
                  'Comment',
                  'stage_name', 'grouping', 'note')

PRODUCTION_HEADER = ('prod_flow', 'ref_direction', 'ref_value', 'ref_unit',
                     'direction', 'balance_yn', 'child_flow', 'units', 'amount', 'amount_hi', 'amount_lo',
                     'descend', 'stage_name',
                     'target_origin', 'target_flow', 'target_name', 'target_ref', 'locale',
                     'add_taps', 'note', 'Name', 'Comment', 'compartment')

TAPS_HEADER = ('tap_recipe', 'flow_origin', 'flow_name_or_ref', 'direction', 'target_origin', 'target_ref', 
               'adjust_value', 'scale_value', 'value_units')

OBSERVATIONS_HEADER = ('activity', 'child_flow', 'scenario', 'parameter', 'units',
                       'anchor_origin', 'anchor', 'anchor_flow', 'descend', 'comment')


HEADERS = {
    'quantities': QUANTITIES_HEADER,
    'flows': FLOWS_HEADER,
    'flowproperties': FLOWPROPERTIES_HEADER,
    'spanners': SPANNERS_META,
    'spanner': SPANNER_HEADER,
    'production': PRODUCTION_HEADER,
    'taps': TAPS_HEADER,
    'observations': OBSERVATIONS_HEADER
}


WIDTHS = {
    'production': {'prod_flow': 220,
                   'child_flow': 170,
                   'direction': 69,
                   'units': 69,
                   'descend': 65,
                   'amount': 75,
                   'amount_hi': 75,
                   'amount_lo': 75
                   },
    'flows': {'external_ref': 220,
              'referenceQuantity': 140,
              'Name': 300},
    'quantities': {'external_ref': 220,
                   'referenceUnit': 140,
                   'Name': 300},
    'flowproperties': {'flow': 220},
    'observations': {'activity': 220,
                     'child_flow': 220,
                     'units': 69},
    'taps': {'tap_recipe': 200},
    'spanners': {'external_ref': 150,
                 'name': 220,
                 'author': 160},
    'spanner': {'flow': 220,
                'Name': 260,
                'units': 69,
                'amount': 75,
                'amount_hi': 65,
                'amount_lo': 65,
                'context': 60,
                'descend': 60
                }
    }


_flows = _dark_magenta = [k/256 for k in (0xcb, 0, 0xcb)]
_model = _dark_green = [k/256 for k in (0x21, 0xa2, 0x60)]
_study = _cornflower_blue = [k/256 for k in (0x64, 0x95, 0xed)]


TAB_COLORS = {
    'quantities': _flows,
    'flows': _flows,
    'flowproperties': _flows,
    'spanners': _model,
    'spanner': _model,
    'production': _study,
    'taps': _study,
    'observations': _study
}



from antelope_foreground.foreground_catalog import NoSuchForeground

from antelope_core.contexts import NullContext
from antelope import ConversionError, comp_dir
from antelope import EntityNotFound, enum, MultipleReferences

from .observations_from_spreadsheet import ObservationsFromSpreadsheet

from .exchanges_from_spreadsheet import exchanges_from_spreadsheet

import re
import logging
from collections import defaultdict

tr = str.maketrans(' ', '_', ',[]()*&^%$#@/\\')


class AmbiguousResult(Exception):
    """
    Raised when there are several choices after filtering, and strict is True
    """
    pass


class MissingValue(Exception):
    """
    Used when a quick-link spec is missing a value and no parent is provided
    """
    pass


def _flow_to_ref(name):
    n = name.translate(tr).lower()
    if n.startswith('flow_'):
        fl = n
    else:
        fl = 'flow_' + n
    return fl


class TapSpecError(Exception):
    pass


class TapSpec:
    def __init__(self, child_flow, child_direction, target, term_flow=None,
                 adjust_value=None, scale_value=None, value_units=None, **kwargs):
        """

        :param child_flow:
        :param child_direction:
        :param target:
        :param adjust_value:
        :param scale_value:
        :param value_units:
        """
        self.child_flow = child_flow
        self.child_direction = child_direction
        if adjust_value is not None and scale_value is not None:
            print('ignoring adjust_value in favor of scale_value')
            adjust_value = None
        if scale_value is not None and value_units is not None:
            print('ignoring value_units for scale_value')
            value_units = None

        self.target = target
        self.term_flow = term_flow
        self.adjust_value = adjust_value
        self.scale_value = scale_value
        self.value_units = value_units
        self.args = kwargs

    @property
    def key(self):
        """
        A tap can only be uniquely defined by its flow and direction
        :return:
        """
        return self.child_flow, self.child_direction


class QuickAndEasy(object):
    """
    This is a basic support class for building operable, linked models from basic specifications provided by the
    user.  It assumes that modeling information is stored in an xlrd-like document, canonically a google sheets
    spreadsheet is preferred because it is so easy to both read and write content on-the-fly*.

    The class includes the following features:
     - fg - an Antelope foreground
     - xlsx - a spreadsheet conforming with the antelope_core.archives.xlsx_updater.XlsxUpdater spec
     - terms - a dictionary of keywords to pre-defined anchors that are used in the model
     - find_background_rx() - a heuristic pathway to convert an open-ended anchor specification into a reference flow
     - new_link() - a function to create a new link in a fragment tree
     - to_background() - a utility function that finds an anchor for a fragment link (invokes find_background_rx())
     - add_tap() - a fundamental feature that converts a background exchange into an observable foreground exchange
     - load_process_model() - creates a model from a spreadsheet specification using exchanges_from_spreadsheet


    More features are added in the ModelMaker subclass.


    * - Writing to the document is NOT required in any of the core functions, however it is useful when the XLSX
     document contains parameters and scenarios that may be altered programmatically.

     In principle it should also be easy to write to e.g. openpyxl spreadsheets, but it is not the case because
     those spreadsheets have explicitly separate in-memory and on-disk versions.  To achieve the same functionality
     as we get with the google doc, we would have to write the change in-memory, save the change to disk, and then
     re-load the file.  That is all upstream in xlstools; for now we are sticking with google docs only.
    """

    def _get_one(self, hits, strict=False, **kwargs):
        hits = list(hits)
        if len(hits) == 1:
            return hits[0]
        elif len(hits) == 0:
            raise EntityNotFound
        else:
            _ = enum(hits)
            if strict:
                raise AmbiguousResult('Ambiguous termination: %d results found' % len(hits))
            print('Warning: Ambiguous termination: %d results found' % len(hits))
            return hits[0]

    @classmethod
    def by_name(cls, cat, fg_name, terms=None, **kwargs):
        """
        NOTE: this resets the foreground.  this is a bit foolish given how badly we handle reset foregrounds.
        :param cat:
        :param fg_name:
        :param terms:
        :param kwargs:
        :return:
        """
        try:
            fg = cat.foreground(fg_name, reset=True)
        except NoSuchForeground:
            fg = cat.create_foreground(fg_name)
        return cls(fg, terms=terms, **kwargs)

    def set_terms(self, terms):
        """
        Maps flow / term specifications (string) to termination specs.
        :param terms: should be a mapping. Keys should be external_ref or term external_ref or context name. Values
         can be: an origin-external ref tuple, an entity (with entity_type attribute), or None (force cutoff).
        :return:
        """
        if terms:
            for k, v in terms.items():
                if isinstance(v, tuple) or isinstance(v, list):
                    self._terms[k] = self.fg.catalog_ref(*v)
                elif v is None:
                    self._terms[k] = None
                elif hasattr(v, 'entity_type'):
                    self._terms[k] = v
                else:
                    print('skipping unrecognized term spec %s: %s' % (k, v))

    def _populate_unit_map(self):
        for q in ('mass', 'volume', 'net calorific value', 'number of items', 'node activity',
                  'length', 'area', 'freight', 'vehicle transport', 'person transport'):  # last one stated wins
            q_can = self.fg.get_canonical(q)
            self.add_to_unit_map(q_can)
        kwh = self.fg.get_canonical('kWh')
        for u in ('kWh', 'MWh', 'GWh'):
            try:
                kwh.convert(u)
                self._unit_map[u] = kwh.external_ref
            except ConversionError:
                pass

    def add_to_unit_map(self, quantity):
        for u in quantity['unitconversion'].keys():
            self._unit_map[u] = quantity.external_ref

    def __init__(self, fg, terms=None, xlsx=None, quiet=True, taps=None):
        """
        A quick-and-easy model builder.  Pass in a foreground to work with, a dictionary of terms mapping nickname to
        origin + external ref, and an optional XlrdLike spreadsheet
        :param fg:
        :param terms: (k, (v,...)). Creates a map of keyname k to CatalogRef(*v). Used to terminate inventory flows.
        :param xlsx:
        :param quiet: passed to fg archive builder
        :param taps: name of spreadsheet containing flow taps (flow origin + ref tapped to target origin + ref)
        :param
        """
        self._fg = fg
        self._terms = {}
        self._xlsx = None
        self._quiet = quiet
        self.set_terms(terms)
        self._unit_map = dict()
        self._populate_unit_map()
        self._taps = defaultdict(dict)

        if xlsx:
            self.xlsx = xlsx

        if taps:
            self.load_taps_from_spreadsheet(taps)

    @property
    def xlsx(self):
        return self._xlsx

    @xlsx.setter
    def xlsx(self, xlsx):
        """
        Automatically loads quantities, flows, and flowproperties
        :param xlsx:
        :return:
        """
        if xlsx:
            self._xlsx = xlsx
            self.refresh_entities()

    @property
    def quantities(self):
        """
        generates quantities by entry in 'quantities'
        :return:
        """
        quantities = self._xlsx.sheet_by_name('quantities')
        for n in range(1, quantities.nrows):
            d = quantities.row_dict(n)
            if d.get('external_ref') is None:
                continue
            if self.fg[d['external_ref']] is None:
                continue
            yield self.fg[d['external_ref']]

    def refresh_entities(self, quiet=None):
        if self._xlsx is None:
            raise ValueError('xlsx is missing')
        if quiet is None:
            quiet = self._quiet
        try:
            self.fg.apply_xlsx(self._xlsx, quiet=quiet)
        except AttributeError:
            next(self.fg._iface('foreground')).apply_xlsx(self._xlsx, quiet=quiet)
        for q in self.quantities:
            self.add_to_unit_map(q)

    @property
    def fg(self):
        return self._fg

    def __getitem__(self, item):
        e = self.fg.__getitem__(item)
        if e is None:
            # the buck has to stop somewhere
            raise KeyError(item)
        return e

    def terms(self, term):
        return self._terms[term]

    def get_flow_by_name_or_ref(self, origin, flow_name_or_ref, strict=True):
        query = self.fg.cascade(origin)
        try:
            flow = query.get(flow_name_or_ref)
        except EntityNotFound:
            flows = filter(lambda x: not self.fg.get_context(x.context).elementary,
                           query.flows(Name='^%s$' % flow_name_or_ref))
            flow = self._get_one(flows, strict=strict)
        return flow

    def find_background_rx(self, origin, external_ref=None, process_name=None, flow_name_or_ref=None, strict=True,
                           **kwargs):
        """
        The purpose of this is to retrieve a unique termination from a user specification. 
        Order of preference here is as follows:
        if external_ref is supplied, just get the straight catalog ref: origin + external_ref

        otherwise if process_name is supplied, search for it by name (regex) with filters; return rx matching flow_name
        otherwise, search for flow_name in origin, filtering by non-elementary context, then find targets with filtering
        (this includes fragments_with_flow() if the origin is a foreground)
        
        :param origin: 
        :param external_ref: 
        :param process_name: exact name of background process
        :param flow_name_or_ref: exact flow name OR external ref of flow
        :param strict: [True] If true, raise an AmbiguousResult exception if multiple hits are found; if False, go
         ahead and use the "first" (which is nondetermininstic).  Provide kwargs to filter.
        :param kwargs: 
        :return: 
        """
        if external_ref:
            term = self.fg.cascade(origin).get(external_ref)
        else:
            query = self.fg.cascade(origin)
            if process_name:
                try:
                    term = self._get_one(query.processes(Name='^%s$' % process_name, **kwargs), strict=strict)
                except EntityNotFound:
                    term = self._get_one(query.processes(Name='^%s' % process_name, **kwargs), strict=strict)
            elif flow_name_or_ref:
                flow = self.get_flow_by_name_or_ref(origin, flow_name_or_ref, strict=strict)

                if hasattr(query, 'fragments_with_flow'):
                    term = self._get_one(query.fragments_with_flow(flow, reference=True))
                else:
                    processes = flow.targets()
                    for k, v in kwargs.items():
                        if v is None:
                            continue
                        processes = list(filter(lambda x: bool(re.search(v, x.get(k), flags=re.I)), processes))
                    term = self._get_one(processes, strict=strict)
            else:
                raise AmbiguousResult('Either process_name or flow_name must be provided')

        try:
            return term.reference()
        except MultipleReferences:
            return term.reference(flow_name_or_ref)

    def _new_reference_fragment(self, flow, direction, external_ref):
        frag = self.fg[external_ref]
        if frag is None:
            frag = self.fg.new_fragment(flow, direction, external_ref=external_ref)

        return frag

    # this is from the misbegotten demo
    def new_link(self, flow_name, ref_quantity, direction, amount=None, units=None, flow_ref=None, parent=None, name=None,
                 stage=None,
                 prefix='frag',
                 balance=None):
        """
        Just discovered that 'balance' is actually a direction

        am I writing fragment_from_exchanges *again*? this is the API, this function right here

        NO
        the api is fragment_from_exchange. and yes, i am writing it again.

        The policy of this impl. is to create from scratch.  no need to re-run + correct: just scratch and throw out

        :param flow_name:
        :param ref_quantity: of flow
        :param direction: of fragment
        :param amount:
        :param units:
        :param flow_ref:
        :param parent:
        :param name:
        :param stage:
        :param prefix: what to add to the auto-name in order to
        :param balance: direction='balance' should be equivalent;; direction is irrelevant under balance
        :return:
        """
        if flow_ref is None:
            flow_ref = _flow_to_ref(flow_name)

        rq = self.fg.get_canonical(ref_quantity)

        flow = self.fg.add_or_retrieve(flow_ref, rq, flow_name)

        external_ref = name or None
        if parent is None:
            if flow_ref.startswith('flow_'):
                auto_name = flow_ref[5:]
            else:
                auto_name = '%s_%s' % (prefix, flow_ref)
            external_ref = external_ref or auto_name

            frag = self._new_reference_fragment(flow, direction, external_ref)
            self.fg.observe(frag, exchange_value=amount, units=units)
        else:
            if direction == 'balance':
                balance = True
                direction = 'Input'  # WLOG
            if balance:
                frag = self.fg.new_fragment(flow, direction, parent=parent, balance=True)
            else:
                frag = self.fg.new_fragment(flow, direction, value=1.0, parent=parent, external_ref=external_ref)
                self.fg.observe(frag, exchange_value=amount, units=units)

        if stage:
            frag['StageName'] = stage

        return frag

    '''
    def to_background(self, bg, origin, external_ref=None, process_name=None, flow_name=None, locale=None,
                      scenario=None,
                      scaleup=1.0,
                      **kwargs):
        """
        This takes a new link and terminates it via balance to a background node that is retrieved by search
        :param bg:
        :param origin:
        :param external_ref:
        :param process_name: exact name of process
        :param flow_name: exact name or ref of flow
        :param locale:
        :param scenario:
        :param scaleup:
        :return:
        """
        kwargs.update({'process_name': process_name, 'flow_name_or_ref': flow_name, 'external_ref': external_ref})
        if locale:
            kwargs['SpatialScope'] = locale

        rx = self.find_background_rx(origin, **kwargs)

        child = bg.balance_flow
        if child is None:
            child = self.fg.new_fragment(bg.flow, bg.direction, parent=bg, balance=True,
                                         StageName=bg['StageName'])

        child.clear_termination(scenario)
        child.terminate(rx.process, term_flow=rx.flow, scenario=scenario)
        if scaleup != 1.0:
            scaleup_adj = scaleup - 1.0
            try:
                z = next(k for k in bg.children_with_flow(bg.flow) if k is not child)
            except StopIteration:
                z = self.fg.new_fragment(bg.flow, comp_dir(bg.direction), parent=bg)
            ev = bg.exchange_value(scenario, observed=True) * scaleup_adj
            self.fg.observe(z, exchange_value=ev, scenario=scenario)
            z.terminate(NullContext)  # truncates
    '''

    def add_tap(self, parent, tap_spec: TapSpec, scenario=None,
                include_zero=False, invert_direction=None, **kwargs):
        """
        Use fragment traversal to override an exchange belonging to a terminal activity.
         - retrieve the termination for the appropriate scenario
         - compute the exchange relation for the specified flow exchanged with the terminal node
         - add or retrieve a child flow with the specified flow
         - observe the child flow to have the same exchange value as the computed exchange
         - optionally, terminate the child flow to the designated termination (or to foreground)

        Two notes: 1- it is not appropriate to specify "terminate the flow to its current anchor" because in general
        processes may have multiple instances of the same flow, set to different anchors.  User must specify.

        2- The correct tap computation depends on the implementation of lci(), which matches fragment child flows to
        the process's dependencies.  These dependencies are always reported with positive exchange values in their
        natural direction (waste flows will appear as outputs and not negative-valued inputs). For this reason, the
        tap machinery inverts the direction of taps when the anchor's reference value is negative.  Override this
        by specifying invert_direction=True or False explicitly.  NOTE: this is reinforfced by treatment in the
        TarjanBackground - which reverses the directions of dependencies in the same condition.

        2a- this deals with the case of dependencies *from* [ecoinvent-style] treatment processes
        (negative reference value) but it doesn't help in the case where the dependency *itself* is negative-valued.
        This can't be detected automatically- so when adding taps in the ecoinvent-style, the direction of the
        *exchange relation* query must always be 'Input'.

        For now, we deal with this just by checking to see if 'ecoinvent' is in the query origin, and if so we

        :param parent:
        :param child_flow:
        :param direction: ['Input'] exchange direction w/r/t parent
        :param scenario:
        :param tap_spec: what to terminate the child flow to. None = cutoff. True = to foreground. all others = as specified
        :param term_flow:
        :param include_zero: [False] whether to add and include child flows with observed 0 EVs
        :param invert_direction: [None] whether to toggle the direction of the child flow. Default behavior is to
         toggle direction ONLY IF the anchor's reference value is negative.
        :param kwargs: passed to new fragment creation
        :return:
        """
        if parent.termination(scenario).is_fg and parent.balance_flow:
            parent = parent.balance_flow

        t = parent.termination(scenario)
        _is_ecoinvent = bool(t.term_node.origin.find('ecoinvent') >= 0)

        if tap_spec.adjust_value is not None:
            # override include_zero
            include_zero = True  # if adj-value is set to 0 we obey it
            # this is super hacky (through and through)
            ev = tap_spec.adjust_value
            ev_units = tap_spec.value_units
        else:
            ev_units = None
            if _is_ecoinvent:
                ev = t.term_node.exchange_relation(t.term_flow, tap_spec.child_flow, 'Input')
                if tap_spec.child_direction == 'Output':
                    ev *= -1
            else:
                ev = t.term_node.exchange_relation(t.term_flow, tap_spec.child_flow, tap_spec.child_direction)
            if tap_spec.scale_value:
                ev *= tap_spec.scale_value

        if ev == 0:
            if not include_zero:
                logging.info('Child child_flow returned 0 exchange', parent, tap_spec.child_flow)
                return None

        # tri-state logic
        if invert_direction is True:
            tgt_dir = comp_dir(tap_spec.child_direction)
        elif invert_direction is False:
            tgt_dir = tap_spec.child_direction
        else:
            rv = t.term_node.reference_value(t.term_flow)
            if rv < 0:
                print('Changing direction for inverted reference activity')
                tgt_dir = comp_dir(tap_spec.child_direction)
            else:
                tgt_dir = tap_spec.child_direction

        try:
            c = next(parent.children_with_flow(tap_spec.child_flow, direction=tgt_dir))
        except StopIteration:
            c = self.fg.new_fragment(tap_spec.child_flow, tgt_dir, parent=parent, **kwargs)
        self.fg.observe(c, exchange_value=ev, scenario=scenario, units=ev_units)
        if tap_spec.target is not None:
            if tap_spec.target is True:
                c.terminate(NullContext)
            else:
                c.terminate(tap_spec.target, term_flow=tap_spec.term_flow, scenario=scenario)

        return c

    def apply_tap_recipes(self, node, recipe=True):
        """
        This should
        :param node:
        :param recipe:
        :return:
        """
        for tap_spec in self._taps[recipe].values():
            self.add_tap(node, tap_spec, include_zero=False)

    def store_tap_recipe(self, recipe, tap_spec):
        if recipe is None:
            recipe = True
        self._taps[recipe][tap_spec.key] = tap_spec

    def load_taps_from_spreadsheet(self, sheetname='taps'):
        sheet = self.xlsx[sheetname]
        for r in range(1, sheet.nrows):
            row = sheet.row_dict(r)
            recipe = row.pop('tap_recipe', True)
            org = row.pop('flow_origin', None)
            name_ref = row.pop('flow_name_or_ref', None)
            if name_ref is None:
                print('skipping empty tap row %d' % (r+1))
                continue
            try:
                if org is None:
                    flow = self.fg.get(name_ref)
                else:
                    flow = self.get_flow_by_name_or_ref(org, name_ref, strict=True)
            except (KeyError, AmbiguousResult) as e:
                print('Loading tap from row %d: %s (%s) - skipping' % (r+1, e.__class__.__name__, e.args[0]))
                continue
            tgt_origin = row.pop('target_origin', 'here')
            tgt_ref = row.pop('target_ref', None)
            if tgt_ref is None:
                tgt = NullContext
            else:
                if tgt_origin == 'here' or tgt_origin is None:
                    tgt = self.fg.get(tgt_ref)
                else:
                    tgt = self.fg.cascade(tgt_origin).get(tgt_ref)
            direction = row.pop('direction', 'Input')

            if None in row:
                row.pop(None)
            tap_spec = TapSpec(flow, direction, tgt, **row)

            self.store_tap_recipe(recipe, tap_spec)

    def load_process_model(self, sheetname, prefix=None, auto_anchor=True, include_elementary=True, **kwargs):
        """


        :param sheetname: the sheet containing inventory data
        :param prefix: prefix to prepend (with '_') to the sheetname to form the external ref
        :param auto_anchor: [True] pass to fragment_from_exchanges
        :param include_elementary: [True] pass to fragment_from_exchanges
        :param kwargs: key-value pairs to assign as properties to the parent fragment
        :return:
        """
        sheet = self.xlsx[sheetname]
        if sheet.nrows <= 1:
            print('No exchanges')
            return

        if prefix:
            ref = '%s_%s' % (prefix, sheetname)
        else:
            ref = sheetname

        exch_gen = exchanges_from_spreadsheet(sheet, origin=self.fg.origin)
        parent = self.fg[ref]  # this is BACKWARDS from standard- .get is supposed to silently return None !!MAJOR ALERT
        if parent is None:
            fproc = self.fg.fragment_from_exchanges(exch_gen, ref=ref, term_dict=self._terms,
                                                    include_elementary=include_elementary,
                                                    auto_anchor=auto_anchor)
        else:
            next(exch_gen)  # nowhere do we apply the incoming exch_gen to the parent!?
            fproc = self.fg.fragment_from_exchanges(exch_gen, parent=parent, term_dict=self._terms,
                                                    include_elementary=include_elementary,
                                                    auto_anchor=auto_anchor)

        fproc['StageName'] = sheetname
        for k, v in kwargs.items():
            fproc[k] = v

        fproc.show_tree(True)
        return fproc

    def apply_observations(self, sheetname='observations'):
        scs = self.xlsx[sheetname]
        with ObservationsFromSpreadsheet(self.fg, scs) as obs:
            obs.apply()



"""
A barebones LCA study environment.
The principle here is that a "study" object should be ephemeral, but it should be built from a persistent
collection of "models", which are the building blocks.
"""
from antelope import EntityNotFound, q_node_activity, comp_dir, comp_sense


class DuplicateRoute(Exception):
    """
    for use when a route already exists
    """
    pass


class MarketRequiresParent(Exception):
    pass


class LcaStudy:
    @property
    def _act(self):
        return q_node_activity(self._fg)

    def new_activity_flow(self, name, external_ref=None, **kwargs):
        """
        This should really be a foreground method
        :param name:
        :param external_ref:
        :param kwargs:
        :return:
        """
        if external_ref is None:
            fu = self._fg.new_flow(name, self._act, **kwargs)
        else:
            fu = self._fg.add_or_retrieve(external_ref, self._act, name, strict=True, **kwargs)
        return fu

    def __init__(self, foreground, models=None, reference_flow='LCA Study',
                 study_container='study_container'):
        self._fg = foreground
        if models is None:
            models = foreground
        self._models = models

        self._ref = reference_flow

        self._study = study_container

        self._phonebook = dict()

        self.route_debug = False

    def _resolve_term(self, term):
        """
        Grab an entry from the foreground[s] associated with the project: first _fg, then _models, then _data if
        it exists. Store successful lookups in a "phone book" to cut down on queries doomed to 404
        :param term:
        :return:
        """
        if hasattr(term, 'entity_type'):
            return term
        if term in self._phonebook:
            return self._phonebook[term]
        try:
            t = self._fg.get(term)
            self._phonebook[term] = t
            return t
        except EntityNotFound:
            t = self._models.get(term)
            self._phonebook[term] = t
            return t

    def _matching_foreground(self, origin):
        if origin == self._fg.origin:
            return self._fg
        elif origin == self._models.origin:
            return self._models
        raise KeyError(origin)

    @property
    def reference_flow(self):
        return self.new_activity_flow(self._ref, external_ref='reference_flow')

    @property
    def fg(self):
        return self._fg

    @property
    def models(self):
        return self._models

    @property
    def study_container(self):
        """
        Putting this in a property allows it to be overridden in a subclass
        :return:
        """
        try:
            return self._fg.get(self._study)
        except EntityNotFound:
            study = self._fg.new_fragment(self.reference_flow, 'Output', Name=self._ref)
            self._fg.observe(study, name=self._study)
        return self._fg.get(self._study)

    '''
    Route-building machinery

    This is a bit of a Kluge, but here is the idea:

    A ROUTE SPECIFICATION consists of a sequence of STAGES (as a tuple).  Each STAGE is EITHER:
     - a fragment present in the models foreground, OR
     - a BRANCH.  Any node in a route specification may generate multiple child flows, known as a BRANCH, and
       designated with a tuple of fragment specs.  Each subsequent stage must be a tuple with the same length.  
       The first stage may not be a branch.
     - Branches can be nested by including tuples of tuples. 

    For each stage, the route builder will retrieve the target and create a fragment whose flow matches the target's.
    It will then terminate the fragment to the target.  That stage then becomes the parent of the subsequent stage.

    When a branch is encountered, it will create multiple child flows from the parent, one for each entry in the branch.
    EVERY SUBSEQUENT STAGE must have the same number of entries as the branch, and these entries are appended as
    child nodes to the branched parents.  Use None to end a branch (None must be supplied as a placeholder throughout
    the route).

    '''

    def _make_single_link(self, parent, direction, child, stage_name=None):
        if self.route_debug:
            print('_make_single_link: %s, %s, %s' % (parent, direction, child))
        try:
            term = self.models.get(child)
        except EntityNotFound:
            term = self.fg.get(child)
        ''' # not sure how to use this yet
        if term.entity_type == 'flow':
            # cutoff
            return term
        '''
        if parent is None:
            # 'comp_dir' because new_fragment UX expects input w.r.t. fragment for reference flows
            f = self.fg.new_fragment(term.flow, comp_dir(direction))
        else:
            if isinstance(parent, list) and len(parent) == 1:
                print('found a list')
                parent = parent[0]
            f = self.fg.new_fragment(term.flow, term.direction, parent=parent)  # we don't *know* the child flow dirn
            # the sign change happens at either of 2 pts: on flow matching in subfrag traversal; or at node_inbound_ev
        f.terminate(term)
        if stage_name:
            f['StageName'] = stage_name
            f.term.descend = False
        return f

    def _make_link_or_links(self, parent, direction, child_or_children, stage_names):
        """
        Ahh, recursion
        :param parent:
        :param direction:
        :param child_or_children:
        :param stage_names:
        :return:
        """
        if self.route_debug:
            print('_make_link_or_links: %s, %s, %s' % (parent, direction, child_or_children))
        if isinstance(child_or_children, dict):
            if isinstance(parent, list):
                # can't imagine this coming up, but let's let it ride
                return [self._make_link_or_links(k, direction, child_or_children, stage_names) for k in parent]
            else:
                obs = self.models[child_or_children['child_flow']]
                if parent is None:
                    mkt_node = self.fg.new_fragment(obs, comp_dir(direction))  # is this getting out of hand?
                else:
                    mkt_node = self.fg.new_fragment(obs, direction, parent=parent)

                if self.make_market(mkt_node, child_or_children['market'], sense=comp_sense(direction),
                                    stage_names=stage_names):
                    # sequencing here is going to be a problem- gotta be just sorted by flow.external_ref
                    return sorted(mkt_node.child_flows, key=lambda x: x.flow.external_ref)
                else:
                    return parent  # aha

        if isinstance(child_or_children, tuple):
            f = []
            if isinstance(parent, list):
                for i, p in enumerate(parent):
                    if p is None or child_or_children[i] is None:
                        f.append(None)
                    else:
                        f.append(self._make_link_or_links(p, direction, child_or_children[i], stage_names))
            else:
                for c in child_or_children:
                    f.append(self._make_link_or_links(parent, direction, c, stage_names))
        else:
            stage_name = stage_names.get(child_or_children)
            f = self._make_single_link(parent, direction, child_or_children, stage_name=stage_name)
        return f

    def _make_route(self, route, sense, stage_names, parent=None):
        """
        This can be used internally to construct routes from the models foreground as tuple specifications
        :param route:
        :param sense: whether the route is a Source of inputs to the study system or a sink for Outputs (i.e. whether it
        will get plugged into inputs or outputs of the study system.
        :param stage_names: dict
        :param parent: [None] for new standalone models. Can also be used to attach routes to existing parents.
        :return:
        """
        if self.route_debug:
            print('_make_route: %s, %s, %s' % (route, sense, parent))
        first = None
        direction = comp_dir(sense)
        if isinstance(route, str) or isinstance(route, dict):
            route = (route, )

        for step in route:
            if first is None and isinstance(step, tuple):
                raise ValueError('Model spec cannot begin with a tuple! %s' % step)
            child = self._make_link_or_links(parent, direction, step, stage_names)
            if first is None:
                if isinstance(child, list):
                    first = child[0].parent  # retrieve the market node
                else:
                    first = child
            parent = child
        return first

    def make_route(self, route_name, route_spec, sense='Sink', stage_names=None):
        """
        Provision (source) and Disposition (sink) routes get built from scratch during LCA runtime in LCA study
        foreground.  Components are drawn from the models foreground.

        The "Route specification" is a tuple of stage specifications.  Each entry in the tuple is a stage.  A stage
        can be either a single entry (corresponding to a fragment in models) or a tuple of entries, signifying a
        branch from the prior stage.  Branch patterns must be maintained for all subsequent stages.  Additional
        branches can be obtained by enclosing a tuple within the tuples.  An entry of "None" indicates that the branch
        ends (but Nones must be maintained through the entire length of the spec)

        Example:

        { "route_1" : ('node_1',),  #       ---> [node_1]
          "route_2" : ('node_2a', 'node_2b'),   # ---> [node_2a] ---> [node_2b ]
          "route_3" : ('node_3a', ('node_3b1', 'node_3b2'), (None, 'node_3c2')), #
           #  ---> [node_3a] -+--> [node_3b1]
           #                  \--> [node_3b2] ---> [node_3c2]
          "route_4" : ('node_4a', ('node_4b1', 'node_4b2'), (('node_4c11', 'node_4c12'), 'node_4c2'), (('node_4d11', 'node_4d12'), 'node_4d2')) #
           #  ---> [node_4a] -+--> [node_4b1] -+--> [node_4c11] ---> [node_4d11]
           #                  \                \--> [node_4c12] ---> [node_4d12]
           #                   \-> [node_4b2] ----> [node_4c2] ----> [node_4d2]
        }

        A dict given as a terminal node can be used to pass a market specification for make_market, but it is
        problematic because the market needs a foreground flow from which to make exchange value observations
        (child flows from fragment terms are not observable).

        So: the dict specification for the market looks like this:
        {'child_flow': flow_ext_ref,
         'market': {market spec...}}
        The child flow is created using the given flow ref, and the market spec is passed to make_market.  The
        number of entries in the market indicates the number of child flows for the next stage.  Because order
        is important, the flows are returned **sorted by their external_refs**.

        Note that the routes specified by the market must exist at the time the market is created.

        :param route_name: The external_ref of the route
        :param route_spec: tuple chain, as above
        :param sense: whether the route is a 'Sink' [default] or a 'Source'
        :param stage_names: a mapping of route spec entries to stage names.  If one is found, the fragment containing
         it is given the assigned stage name, and descend is set to False for that link (aggregating the target within
         the named stage).
        :return: the created fragment
        """
        if stage_names is None:
            stage_names = dict()
        if self.fg[route_name] is None:
            route = self._make_route(route_spec, sense, stage_names)
            if route is None:
                print('^^^^^^^^^^^^^^^^^^^^^^^^^^ No route produced for %s\n%s' % (route_name, route_spec))
            elif route.entity_type == 'flow':
                # do nothing?
                pass
            else:
                self.fg.observe(route, name=route_name)
            if route_name in stage_names:
                route['StageName'] = stage_names[route_name]
            return route
        raise DuplicateRoute('Route %s already exists' % route_name)

    def _check_p_map(self, p_map):
        nc = 0
        for k, v in p_map.items():
            try:
                self._resolve_term(k)
            except EntityNotFound:
                print('Product map entry %s not found' % k)
                return False
            if v is None:
                nc += 1
            else:
                try:
                    float(v)
                except (ValueError, AttributeError):
                    print('Key %s has non-floatable value %s' % (k, v))
                    return False
        if nc != 1:
            print('Wrong number of balance flows (%d)' % nc)
            return False
        return True

    def make_market(self, parent_or_flow, p_map, sense='Sink', stage_names=None):
        """
        Another kind of utility function- this one turns dictionaries into market mixers, either over flows or models.
        For this the user must specify the parent fragment.

        Constitutive difference here is that markets are required to have parents, whereas routes are new fragments.

        The technical reason is that there must be an explicit node to compute the balance.

        Supplying a flow will create a self-standing market for the flow

        The default naming for market knobs provides for autoincrement

        :param parent_or_flow: either a parent fragment whose flow is getting mixed, or a flow to build a self-standing mixer
        :param p_map:
        :param sense:
        :param stage_names:
        :return:
        """
        if self.route_debug:
            print('make_market: %s, %s, %s' % (parent_or_flow, p_map, sense))
        if stage_names is None:
            stage_names = dict()
        if isinstance(p_map, str):
            p_map = {p_map: None}
        direction = comp_dir(sense)
        if parent_or_flow is None:
            raise MarketRequiresParent('Cannot build a free-standing market mixer- parent cannot be None')
        if self._check_p_map(p_map):
            if parent_or_flow.entity_type == 'flow':
                market_ref = 'mix-%s-%s' % (sense, parent_or_flow.external_ref)
                market_ref = market_ref.replace('/', '_')
                parent = self.fg[market_ref]
                if parent is not None:
                    return parent  # don't support dynamic modification of markets

                parent = self.fg.new_fragment(parent_or_flow, comp_dir(direction))
                self.fg.observe(parent, name=market_ref)

            elif parent_or_flow.entity_type == 'fragment':
                parent = parent_or_flow
            else:
                raise TypeError('Unrecognized parent type %s: %s' % (parent_or_flow.entity_type, parent_or_flow.external_ref))
            for k, v in p_map.items():
                term = self._resolve_term(k)  # no need to try-catch this bc we've already checked the p_map
                if term.entity_type == 'flow':
                    flow = term

                elif term.entity_type == 'fragment':
                    flow = parent.flow  # stick with market parent for conservation purposes

                else:
                    raise TypeError(term, "p_map key looks up to wrong or nothing")

                if v is None:
                    c = self.fg.new_fragment(flow, direction, parent=parent, name=term['name'], balance=True)
                else:
                    c = self.fg.new_fragment(flow, direction, parent=parent, name=term['name'])
                    if parent.external_ref == parent.uuid:
                        mix_name = 'mix-%s-%s-%s' % (sense, parent.flow.external_ref, k)
                    else:
                        mix_name = '%s-%s' % (parent.external_ref, k)
                    mix_name = mix_name.replace('/', '_')
                    self.fg.observe(c, name=mix_name, exchange_value=v, auto=True)

                if term.entity_type == 'fragment':
                    stage_name = stage_names.get(k)
                    c.terminate(term, term_flow=term.flow)
                    if stage_name:
                        c.term.descend = False
                        c['StageName'] = stage_name

                    # flows will show up as cut-offs / and/or get forwarded up the superstructure

                # not sure what this is for..
                # if term.external_ref == 'frag-t_waste':
                #     c.term.descend = False
                #     c['StageName'] = 'Waste to Landfill'

            return parent
        print('Not building product map for %s' % parent_or_flow)
        return False

    '''
    Model populating methods
    '''
    def make_routes(self, routes, sense='Sink', stage_names=None):
        for k, v in routes.items():
            try:
                self.make_route(k, v, sense=sense, stage_names=stage_names)
            except DuplicateRoute:
                print('Route %s exists; NOT updating' % k)
                pass
            except EntityNotFound as e:
                print('Route %s: entity not found %s: skipping' % (k, e.args))

    def apply_ad_hoc_parameter(self, adhoc_scenario, param_spec, factor, mult=True):
        """
        Apply an ad hoc parameterization to a uniquely specified child fragment.  User must specify:
         - the name or external ref of the parent fragment
         - the external ref of the child flow

        User may optionally specify:
         - the observed scenario that should be altered [default is base case]

        The routine will find the *first* child flow, retrieve its observed exchange value, multiply it by the factor,
        and enter a new observation under the adhoc_scenario.

        If 'mult=False' is supplied, then the parameter is applied as-is, without multiplication (in this case, any
        reference scenario specification is ignored)

        :param adhoc_scenario: the scenario under which the ad hoc parameter will be observed
        :param param_spec: A tuple having 2 or 3 elements:
         2-element: (fragment_ref, flow_ref) using default_fg, default (observed) reference scenario
         3-element: (fragment_ref, flow_ref, scenario) using default_fg [if origin is not found in cat.foregrounds]
        :param factor: The value by which to multiply the base exchange value. If None, the scenario is un-set.
        :param mult: If True (default), the factor is multiplicative and applied to the adhoc_scenario (future:
        applied to fragment multiplicative root param). if false, is applied directly to the adhoc_scenario.
        :return:
        """
        if len(param_spec) == 2:  # (fragment, child_flow)
            frag, child = param_spec
            sc = None
        elif len(param_spec) == 3:  # (origin, fragment, child_flow)
            frag, child, sc = param_spec
        else:
            print('%s: skipping unrecognized ad hoc parameter %s' % (adhoc_scenario, param_spec))
            return
        tgt = self._resolve_term(frag)
        if tgt is None:
            print('%s: Unable to retrieve fragment %s' % (adhoc_scenario, param_spec))
            return
        fg = self._matching_foreground(tgt.origin)
        flow = fg[child]
        cfs = list(tgt.children_with_flow(flow))
        if len(cfs) == 0:
            cfs = list(tgt.children_with_flow(flow, recurse=True))
            if len(cfs) == 0:
                cfs = list(tgt.children_with_flow(flow, match=True, recurse=True))
        if len(cfs) == 1:
            cf = cfs[0]
            if mult and (factor is not None):
                base_value = cf.exchange_value(scenario=sc, observed=True)
                value = base_value * factor
            else:
                value = factor
            fg.observe(cf, value, scenario=adhoc_scenario)  # observe None should un-set (this is testable!)
        elif len(cfs) == 0:
            print('%s: no child flow found %s' % (adhoc_scenario, param_spec))
        else:
            print('%s: too many (%d) child flows found %s' % (adhoc_scenario, len(cfs), child))
            print('Or, actually, maybe the thing to do is to paramaterize ALL of them!')

    def unset_scenario_knobs(self, scenarios):
        """
        Remove parameter specifications supplied by a scenario.
        Uses the same dictionary structure as set_scenario_knobs, but ignores the value and instead removes
        the observation.
        :param scenarios:
        :return:
        """
        if scenarios is None or len(scenarios) == 0:
            return

        for k, vd in scenarios.items():
            if k is None:
                continue
            for i, v in vd.items():
                if isinstance(i, tuple):
                    self.apply_ad_hoc_parameter(k, i, None)
                elif isinstance(i, str):
                    frag = self._resolve_term(i)
                    fg = self._matching_foreground(frag.origin)
                    fg.observe(frag, scenario=k, exchange_value=None)

    def set_scenario_knobs(self, scenarios, mult=True):
        """
        Apply parameter values to a set of "knobs" (fragment names) to define scenarios.
        Note: if "knob name" is a tuple, interpret it as an ad hoc parameterization, specifying the parent fragment,
        the child to parameterize, and the optional scenario case to be altered.  Ad hoc parameters are multiplicative,
        meaning they are applied to the existing value and not interpreted as absolute values.  This means they
        cannot be used to parameterize zero-valued cases.
        ([origin], parent fragment, child flow, [scenario])
        :param scenarios: (dict of dicts) mapping of scenario names to {knob name: value} mappings - a scenario will
         be created for each key, with the corresponding knobs set to spec.  Use 'scenario': True to add scenario
         flags
        :param mult: whether factors in ad-hoc scenarios should be multiplicative [True] or absolute [False]
        :return: None
        """
        if scenarios is None or len(scenarios) == 0:
            return

        for k, vd in scenarios.items():

            if vd is None:
                continue

            for i, v in vd.items():
                if v is True:
                    # valid setting at runtime; nothing to do here
                    continue
                if isinstance(i, tuple):
                    self.apply_ad_hoc_parameter(k, i, v, mult=mult)
                elif isinstance(i, str):
                    frag = self._resolve_term(i)
                    fg = self._matching_foreground(frag.origin)
                    fg.observe(frag, scenario=k, exchange_value=v)
                else:
                    print('%s: Skipping unknown scenario key %s=%g' % (k, i, v))

    def set_knob_scenarios(self, knobs, unset=False, mult=True):
        """
        Apply parameter values to a set of "knobs" to define scenarios.
        Similar to set_scenario_knobs(), except that the structure of the dict is inverted: instead of the
        parameter settings being grouped by scenario and specified by knob, they are grouped by knob and
        specified by scenario.  This routine simply re-packs the specification and calls set_scenario_knobs().

        :param knobs: (dict of dicts) mapping knob name to (scenario: value)
        :param unset: [False] un-set the given knob-scenario mapping (disregards value)
        :param mult: [True] whether ad hoc parameters are multiplicative or absolute-valued (should be per-parameter, tbh)
        :return:
        """
        if knobs is None or len(knobs) == 0:
            return

        scenarios = dict()
        for knob, mapping in knobs.items():
            for scenario, value in mapping.items():
                if scenario not in scenarios:
                    scenarios[scenario] = dict()
                scenarios[scenario][knob] = value
        if unset:
            self.unset_scenario_knobs(scenarios)
        else:
            self.set_scenario_knobs(scenarios, mult=mult)



class ParamManager(object):
    """
    The parameters sheet has the following columns- square brackets indicate optional:
    (origin) (parameter) [flow_name] [flow_unit] [default].... scenario names
    """
    _sheet = None
    _sheetname = None

    _reserved_names = ('origin', 'parameter', 'flow_name', 'flow_unit',  'default')

    def __init__(self, fg, xlsx, sheetname='parameters'):
        self._xlsx = xlsx
        self._fg = fg
        self.select_sheet(sheetname)

    def select_sheet(self, sheetname):
        if sheetname in self._xlsx.sheet_names():
            self._sheetname = sheetname
            self._update()
            print('Found scenarios: %s' % list(self.scenarios))
        else:
            raise KeyError('Sheet %s not found' % sheetname)

    def _update(self):
        self._sheet = self._xlsx[self._sheetname]

    @property
    def scenarios(self):
        org = False
        ext = False
        for k in self._sheet.row(0):
            if k.value in self._reserved_names:
                if k.value == 'origin':
                    org = True
                if k.value == 'parameter':
                    ext = True
                continue
            yield k.value
        if not org:
            raise ValueError("Required column 'origin' not found")
        if not ext:
            raise ValueError("Required column 'parameter' not found")

    def write_parameters(self):
        """
        This destructively writes the current set of stored parameters to the sheet.
        ad-hoc parameters are not supported, but they could be by finding parameterized fragments that are not knobs.
        :return:
        """
        from pandas import DataFrame

        df = DataFrame(self._fg.knobs(param_dict=True))
        self._xlsx.write_dataframe('parameters', df, clear_sheet=True, fillna='', write_index=False)
        self._update()

    def _get_rows(self):
        for r in range(1, self._sheet.nrows):
            row = self._sheet.row_dict(r)
            if 'parameter' not in row:
                continue
            if row.pop('origin', None) != self._fg.origin:
                continue
            yield row

    def _get_knob_or_child_flow(self, row):
        param = row.pop('parameter')
        kn = self._fg[param]
        if kn is None:
            print('Skipping unknown parameter %s/%s' % (self._fg.origin, param))
            return None
        cf = self._fg[row.pop('flow_name', None)]
        if cf:
            try:
                return next(kn.children_with_flow(cf))
            except StopIteration:
                print('Skipping missing child flow %s/%s:%s' % (self._fg.origin, param, cf.external_ref))
                return None
        return kn

    def write_scenario(self, scenario):
        """
        Write the specified column to the spreadsheet
        :param scenario:
        :return:
        """
        if scenario in self.scenarios:
            col = next(i for i, k in enumerate(self._sheet.row(0)) if k.value == scenario)
            data = []
            start_row = 1
        else:
            col = self._sheet.ncols
            data = [scenario]
            start_row = 0
        for r in range(1, self._sheet.nrows):
            row = self._sheet.row_dict(r)
            if row.get('origin') == self._fg.origin:
                kn = self._get_knob_or_child_flow(row)
                if kn is None:
                    data.append(None)
                    continue

                unit = row['flow_unit']
                val = kn.exchange_value(scenario)
                if val == kn.observed_ev:
                    data.append('')
                else:
                    if unit == kn.flow.unit:
                        data.append(val)
                    else:
                        cf = kn.flow.reference_entity.convert(to=unit)
                        data.append(val * cf)
            else:
                data.append(None)
        print('Writing %d items to %s, scenario %s' % (len(list(filter(None, data))) - 1 + start_row, self._sheetname,
                                                       scenario))
        self._xlsx.write_column(self._sheetname, col, data, start_row=start_row)
        self._update()

    def apply_parameters(self):
        # self._fg.clear_scenarios(terminations=False)  # well, that's foolish
        count = 0
        for row in self._get_rows():
            kn = self._get_knob_or_child_flow(row)
            if kn is None:
                continue

            unit = row.pop('flow_unit', None)
            for k, v in row.items():
                if k in self._reserved_names:
                    continue
                if v is None or v == 'NA':
                    continue
                self._fg.observe(kn, exchange_value=float(v), scenario=k,  units=unit)
                count += 1
        print('Applied %d parameter settings from %s' % (count, self._sheetname))

    def apply_scenario(self, scenario):
        """
        Don't clear the whole foreground; instead re-apply the scenario
        :param scenario:
        :return:
        """
        if scenario in self._reserved_names:
            raise ValueError('Invalid scenario name: %s' % scenario)
        count = 0
        for row in self._get_rows():
            kn = self._get_knob_or_child_flow(row)
            if kn is None:
                continue

            unit = row.pop('flow_unit', None)
            val = row.pop(scenario)
            if val is None:
                kn.set_exchange_value(scenario, None)
            else:
                self._fg.observe(kn, exchange_value=val, scenario=scenario, units=unit)
                count += 1
        print('Applied %d parameter settings from %s, scenario %s' % (count, self._sheetname, scenario))



"""
This is the dynamically-generated study used for running results

CONCEPT: This module / package provides tools to "automagically" build a multi-tiered LCA model that has at its core
an MFA model.  The basic structure of the model is as follows:

#-0-...      Study container (X). Terminated to logistics fragment. Mapping MFA flows to life cycle routes that generate impacts.
  #-0-...    Logistics container (Y). Terminated to activity fragment container. Mapping logistics flows to terminal fragments
    #-0-...  Foreground Activity container (A or a_c). Terminates observed LCA model to various foreground activities, and drive
      |      logistics and market routes in the outer containers
      #- LCA models generated from observational data, generating no impacts and producing balance flows and logistics flows
         accessed by scenarios supplied to the activity container

The study container, logistics container, and activity container are all generic.  The MFA models are generated from
fragment queries, and other MFA-like models can be generated on a study-specific basis.

There are a total of three LcForegrounds involved:

The study and logistics containers are generated on-the-fly / at query time.  The activity container is part of a
semi-persistent observational foreground.  The impact drivers are all stored in a separate / curated modeling foreground.

"""

from antelope import EntityNotFound
from .markets import make_routes, build_market_mix


def build_study_foreground(study, lca, study_container='Study Container', logistics_container='Logistics Container',
                           route_map=None, product_flow_map=None, logistics_map=None, strict=True):
    """
    The study foreground consists of three nested container fragments, each of which implements one stage of LCA
    computation.

    The innermost fragment handles processing activities and is stored semi-persistently in the MFA foreground, along
    with the mfa-derived facility models.  This step is handled in generate_mfa_fragment() below.

    The next-outer fragment handles logistics-- maps MFA logistical flows to LCA transport activities. For this,
    every transporter has to be loaded in mfa_config and mapped here to an lca logistics flow.

    The outermost fragment handles the market disposition of the product flows, and requires the market machinery in
    build_market_mix which, janky though it is, should be fire and forget at this point

    The next-level fragment
    processor activity models
    :param study: a foreground to contain the dynamic study
    :param lca: an LcaModel, having both an observational foreground (fg) and a persistent modeling foreground (models)
    :param study_container: Name (external_ref) of the study container fragment X
    :param logistics_container: Name (external_ref) of the logistics container fragment Y
    :param route_map: see .markets.make_routes()
    :param product_flow_map: maps (fg) flows to market mixes of routes
    :param logistics_map: maps (fg) flows to transport processes
    :return: X and Y, the study overall and the logistics map
    """
    if product_flow_map is None:
        product_flow_map = dict()

    if logistics_map is None:
        logistics_map = dict()

    if route_map is None:
        route_map = dict()

    make_routes(study, lca.models, route_map)

    # build node balance flows
    for tup in lca.lca_entities:
        lca.fg.add_or_retrieve(*tup, strict=strict)

    # build functional unit
    f_u = lca.functional_unit

    # build logistics container
    try:
        log = study.get(logistics_container)
    except EntityNotFound:
        log = study.new_fragment(f_u, 'Output', Name=logistics_container)
        for k, v in logistics_map.items():
            try:
                flow = lca.fg.get(k)
            except EntityNotFound:
                continue
            try:
                term = lca.models.get(v)
            except EntityNotFound:
                continue
            study.new_fragment(flow, 'Input', parent=log, StageName='Reverse Logistics').terminate(term, descend=False)
        study.observe(log, name=logistics_container)
        a_c = lca.activity_container
        log.terminate(a_c)

    # build study container
    try:
        frag = study.get(study_container)
    except EntityNotFound:
        frag = study.new_fragment(f_u, 'Output', Name=study_container)
        for k, p_map in product_flow_map.items():
            flow = lca.fg.get(k)
            c = study.new_fragment(flow, 'Output', parent=frag)
            if build_market_mix(study, c, lca.models, p_map):
                print('Built model for %s' % k)
            else:
                print('Skipped badly formed specification for %s' % k)
        study.observe(frag, name=study_container)
        frag.terminate(log)
    return frag, log


def scenario_knobs(knobs, scenarios):
    """
    Apply parameter values to a set of "knobs" (fragment names) to define scenarios.
    :param knobs: mapping of fragment names (external_ref) to fragments
    :param scenarios: mapping of scenario names to knob: value mappings (dict of dicts)
    :return: None
    """
    for k, vd in scenarios.items():

        for i, v in vd.items():
            if i in knobs:
                knobs[i].observe(v, scenario=k)
            elif v is True:
                # valid setting at runtime; nothing to do here
                continue
            else:
                print('%s: Skipping unknown scenario key %s=%g' % (k, i, v))
